import numpy as np
import numpy.typing as npt
from tt_sketch.tensor import DenseTensor
from tt_sketch.utils import matricize


def sketch_omega_dense(
    left_sketch: npt.NDArray,
    rigth_sketch: np.ndarray,
    *,
    tensor: DenseTensor,
    mu: int,
    **kwargs
) -> npt.NDArray:
    X_mat = matricize(tensor.data, range(mu + 1), mat_shape=True)
    return left_sketch @ X_mat @ rigth_sketch.T


def sketch_psi_dense(
    left_sketch: npt.NDArray,
    right_sketch: np.ndarray,
    *,
    tensor: DenseTensor,
    mu: int,
    **kwargs
) -> npt.NDArray:
    ndim = tensor.ndim
    tensor_data = tensor.data
    if left_sketch is None:
        tensor_mat = matricize(tensor_data, 0, mat_shape=True)
        Psi = tensor_mat @ right_sketch.T
        Psi = Psi.reshape((1,) + Psi.shape)
    elif right_sketch is None:
        tensor_mat = matricize(tensor_data, ndim - 1, mat_shape=True).T
        Psi = left_sketch @ tensor_mat
        Psi = Psi.reshape(Psi.shape + (1,))
    else:
        # Compute order 3 tensorisation of X
        tensor_ord3 = matricize(tensor_data, range(mu + 1), mat_shape=False)
        tensor_ord3 = tensor_ord3.reshape(
            np.prod(tensor_ord3.shape[:mu], dtype=int),
            tensor_ord3.shape[mu],
            tensor_ord3.shape[mu + 1],
        )
        Psi = np.einsum(
            "ij,jkl,ml->ikm",
            left_sketch,
            tensor_ord3,
            right_sketch,
            optimize="optimal",
        )
    return Psi



# %%
from itertools import product
import pandas as pd

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from tt_sketch.drm import (
    DenseGaussianDRM,
    TensorTrainDRM,
)
from tt_sketch.tensor import DenseTensor
from tt_sketch.utils import sqrt_tensor

from experiment_base import (
    Experiment,
    experiment_orthogonal_sketch,
    experiment_stream_sketch,
    experiment_tt_svd,
    experiment_hmt_sketch,
)


n_dim = 5
dim = 10
shape = (dim,) * n_dim
tensor = DenseTensor(sqrt_tensor(shape, a=-0.2, b=2)).to_sparse()
csv_filename = "results/sum-sqrt.csv"
experiment = Experiment(csv_filename)
# %%
# %%
ranks = range(1, 16)
drm_types = [
    DenseGaussianDRM,
    TensorTrainDRM,
]
runs = range(20)

for rank, run, drm_type in tqdm(
    list(product(ranks, runs, drm_types)), desc="OTTS"
):
    experiment.do_experiment(
        tensor,
        "OTTS",
        experiment_orthogonal_sketch,
        left_rank=rank,
        right_rank=rank * 2,
        left_drm_type=drm_type,
        right_drm_type=drm_type,
        run=run,
    )

for rank, run, drm_type in tqdm(
    list(product(ranks, runs, drm_types)), desc="HMT"
):
    experiment.do_experiment(
        tensor,
        "HMT",
        experiment_hmt_sketch,
        rank=rank,
        drm_type=drm_type,
        run=run,
    )

for rank, run, drm_type in tqdm(
    list(product(ranks, runs, drm_types)), desc="STTA"
):
    experiment.do_experiment(
        tensor,
        "STTA",
        experiment_stream_sketch,
        left_rank=rank,
        right_rank=rank * 2,
        left_drm_type=drm_type,
        right_drm_type=drm_type,
        run=run,
    )

for rank in tqdm(ranks, desc="TT-SVD"):
    experiment.do_experiment(tensor, "TT-SVD", experiment_tt_svd, rank=rank)


# %%
import matplotlib.pyplot as plt
import pandas as pd

df = pd.read_csv(csv_filename)
df
# %%
plt.figure(figsize=(8, 4))
ttsvd = df[df["name"] == "TT-SVD"]

ssketch = df[df["name"] == "OTTS"]
plot_ranks = ssketch["left_rank"].unique()
plt.plot(plot_ranks, ttsvd.error.values, "-o", label="TT-SVD", ms=3)

# drms = {
#     "DenseGaussianDRM": "OTTS, Gaussian DRM",
#     "TensorTrainDRM": "OTTS, TT-DRM",
# }
# for i, (drm, drm_name) in enumerate(drms.items()):
#     error_gb = (
#         ssketch[ssketch["left_drm_type"] == drm].groupby("left_rank").error
#     )
#     errors05 = error_gb.quantile(0.5).values
#     errors08 = error_gb.quantile(0.8).values - errors05
#     errors02 = errors05 - error_gb.quantile(0.2).values
#     plt.errorbar(
#         plot_ranks - 0.05 * (i + 0.5),
#         errors05,
#         yerr=np.stack([errors02, errors08]),
#         label=drm_name,
#         capsize=3,
#         linestyle="",
#     )

hsketch = df[df["name"] == "HMT"]
drms = {
    "DenseGaussianDRM": "TT-HMT, Gaussian DRM",
    "TensorTrainDRM": "TT-HMT, TT-DRM",
}
for i, (drm, drm_name) in enumerate(drms.items()):
    error_gb = hsketch[hsketch["drm_type"] == drm].groupby("rank").error
    errors05 = error_gb.quantile(0.5).values
    errors08 = error_gb.quantile(0.8).values - errors05
    errors02 = errors05 - error_gb.quantile(0.2).values
    plt.errorbar(
        plot_ranks - 0.15 * (1.5-i),
        errors05,
        yerr=np.stack([errors02, errors08]),
        label=drm_name,
        capsize=3,
        linestyle="",
    )

ssketch = df[df["name"] == "STTA"]
drms = {
    "DenseGaussianDRM": "STTA, Gaussian DRM",
    "TensorTrainDRM": "STTA, TT-DRM",
}
for i, (drm, drm_name) in enumerate(drms.items()):
    error_gb = (
        ssketch[ssketch["left_drm_type"] == drm].groupby("left_rank").error
    )
    errors05 = error_gb.quantile(0.5).values
    errors08 = error_gb.quantile(0.8).values - errors05
    errors02 = errors05 - error_gb.quantile(0.2).values
    plt.errorbar(
        plot_ranks + 0.15 * (i + 0.5),
        errors05,
        yerr=np.stack([errors02, errors08]),
        label=drm_name,
        capsize=3,
        linestyle="",
    )

plt.xticks(ranks)
plt.ylabel("Relative error")
plt.xlabel("TT-rank")
plt.yscale("log")
plt.legend()
plt.title("Approximation of sqaure-root-sum tensor")
plt.savefig("results/plot-sum-sqrt.pdf", transparent=True, bbox_inches="tight")
plt.show()

# %%



# %%
"""In this experiment we compare the effect of different levels of
right-oversampling. On both the median error, and the variance in the error."""

"""As target tensor let's pick something for which we get a decent amount of
variance. For example the sum-of-tt tensor"""

import numpy as np
import pandas as pd
from itertools import product
from tqdm import tqdm
from typing import List
from tt_sketch.drm import TensorTrainDRM
import matplotlib.pyplot as plt
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor

from experiment_base import (
    Experiment,
    experiment_orthogonal_sketch,
    experiment_stream_sketch,
    experiment_tt_svd,
)

n_dims = 5
dim = 10
shape = (dim,) * n_dims

tt_rank = 3
num_tts = 20

np.random.seed(179)
tt_summands_list: List[Tensor] = []
coeffs = np.logspace(0, -20, num_tts)
for coeff in coeffs:
    tensor_tt = TensorTrain.random(shape, rank=tt_rank) * coeff
    tt_summands_list.append(tensor_tt)
tensor = TensorSum(tt_summands_list)
csv_filename = "results/right_oversampling.csv"
experiment = Experiment(csv_filename)

# %%
left_rank = 10
# right_ranks = range(11, 31)
oversampling_params = range(2, 41, 2)

runs = range(100)

# for right_rank, run in tqdm(list(product(right_ranks, runs)), desc="OTTS"):
#     experiment.do_experiment(
#         tensor,
#         "OTTS",
#         experiment_orthogonal_sketch,
#         left_rank=left_rank,
#         right_rank=right_rank,
#         run=run,
#     )

for oversampling, run in tqdm(
    list(product(oversampling_params, runs)), desc="STTA"
):
    experiment.do_experiment(
        tensor,
        "STTA",
        experiment_stream_sketch,
        left_rank=left_rank,
        right_rank=left_rank + oversampling,
        run=run,
    )

experiment.do_experiment(tensor, "tt_svd", experiment_tt_svd, rank=left_rank)
experiment.do_experiment(
    tensor, "tt_svd", experiment_tt_svd, rank=left_rank - 1
)
experiment.do_experiment(
    tensor, "tt_svd", experiment_tt_svd, rank=left_rank - 2
)

# %%
df = pd.read_csv(csv_filename)

plt.figure(figsize=(8, 4))
ttsvd = df[(df["name"] == "tt_svd") & (df["rank"] == left_rank-2)]
# plt.axhline(ttsvd["error"].iloc[0], ls="--", color="k", label="TT-SVD")

# rsketch = df[df["name"] == "OTTS"]
# right_ranks = rsketch["right_rank"].unique()
# plot_ranks = right_ranks - left_rank
# error_gb = rsketch.groupby(rsketch["right_rank"]).error
# errors05 = error_gb.quantile(0.5).values
# plt.plot(plot_ranks, errors05, marker=".", label="OTTS median")
# errors08 = error_gb.quantile(0.8).values
# plt.plot(plot_ranks, errors08, "--", label="OTTS 80th percentile")
# errors02 = error_gb.quantile(0.2).values
# plt.plot(plot_ranks, errors02, "--", label="OTTS 20th percentile")


ssketch = df[df["name"] == "STTA"]
right_ranks = ssketch["right_rank"].unique()
plot_ranks = right_ranks - left_rank
error_gb = ssketch.groupby(ssketch["right_rank"]).error
errors05 = error_gb.quantile(0.5).values
plt.plot(plot_ranks, errors05, marker=".", label="STTA median")
errors08 = error_gb.quantile(0.8).values
plt.plot(plot_ranks, errors08, "-.", label="80th percentile")
errors02 = error_gb.quantile(0.2).values
plt.plot(plot_ranks, errors02, "-.", label="20th percentile")

plt.yscale("log")
plt.xticks(plot_ranks)
plt.xlabel("Right oversampling $\ell$")
plt.ylabel("L2 error")
plt.legend()
plt.title("Effect of right oversampling $\ell$ on approximation error")

plt.savefig(
    "results/plot-right-oversampling.pdf", transparent=True, bbox_inches="tight"
)
plt.show()
# %%



# %%
"""
Simple timings experiment. We compare the time required to sketch a TT of
varying rank. The rank of the sketches also vary.
"""

import numpy as np
import pandas as pd
from experiment_base import (
    Experiment,
    experiment_orthogonal_sketch,
    experiment_stream_sketch,
    experiment_hmt_sketch,
    experiment_tt_svd,
)
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.sketch import (
    SketchedTensorTrain,
    stream_sketch,
    orthogonal_sketch,
    hmt_sketch,
)
from itertools import product

from tt_sketch.tensor import TensorTrain

csv_filename = "results/timings150.csv"

num_runs = 100
runs = list(range(num_runs))
shape = (100,) * 5
experiment = Experiment(csv_filename)
tt_rank = 150
sketch_ranks = np.arange(5, tt_rank + 0.1, 10, dtype=int)
SEED = 179

# %%


def tt_exp_decay(shape, tt_rank, min_svdval=-20, seed=None):
    tensor = TensorTrain.random(
        shape, rank=tt_rank, orthog=True, trim=True, seed=seed
    )
    for i, C in enumerate(tensor.cores):
        C_shape = C.shape
        left_mat_shape = (C_shape[0] * C.shape[1], C_shape[2])
        right_mat_shape = (C_shape[0], C_shape[1] * C.shape[2])
        if min(left_mat_shape) > min(right_mat_shape):
            mat_shape = left_mat_shape
        else:
            mat_shape = right_mat_shape
        C = C.reshape(mat_shape)
        U, S, Vt = np.linalg.svd(C, full_matrices=False)
        S *= np.logspace(0, min_svdval, len(S))
        C = (U @ np.diag(S) @ Vt).reshape(C_shape)
        tensor.cores[i] = C
    return tensor


def tt_error_func(tt1, tt2):
    if isinstance(tt1, SketchedTensorTrain):
        tt1 = tt1.to_tt()
    tt2_norm = tt2.norm()
    if tt2_norm == 0:
        return np.infa
    error = tt1.add(-tt2).norm() / tt2_norm
    return error


tensors = []
for i in tqdm(runs, desc="creating target tensors"):
    tensor = tt_exp_decay(shape, min_svdval=-10, tt_rank=tt_rank, seed=SEED + i)
    tensors.append(tensor)

# %%
# for tt_rank, sketch_rank, run in tqdm(
#     list(product(tt_ranks, sketch_ranks, runs)), desc="OTTS"
# ):
#     # tensor = TensorTrain.random(shape, rank=tt_rank)
#     tensor = tt_exp_decay(shape, tt_rank, seed=SEED+run)
#     experiment.do_experiment(
#         tensor,
#         "OTTS",
#         experiment_orthogonal_sketch,
#         left_rank=sketch_rank,
#         right_rank=sketch_rank * 2,
#         tensor_rank=tt_rank,
#         sketch_rank=sketch_rank,
#         run=run,
#     )


for run, sketch_rank in tqdm(list(product(runs, sketch_ranks)), desc="STTAx2"):
    # tensor = TensorTrain.random(shape, rank=tt_rank)
    # tensor = tt_exp_decay(shape, tt_rank, seed=SEED + run)
    tensor = tensors[run]
    experiment.do_experiment(
        tensor,
        "STTAx2",
        experiment_stream_sketch,
        left_rank=sketch_rank,
        right_rank=sketch_rank * 2,
        tensor_rank=tt_rank,
        sketch_rank=sketch_rank,
        run=run,
        error_func=tt_error_func,
    )

for run, sketch_rank in tqdm(list(product(runs, sketch_ranks)), desc="STTA+3"):
    # tensor = TensorTrain.random(shape, rank=tt_rank)
    # tensor = tt_exp_decay(shape, tt_rank, seed=SEED + run)
    tensor = tensors[run]
    experiment.do_experiment(
        tensor,
        "STTA+3",
        experiment_stream_sketch,
        left_rank=sketch_rank,
        right_rank=sketch_rank + 3,
        tensor_rank=tt_rank,
        sketch_rank=sketch_rank,
        run=run,
        error_func=tt_error_func,
    )

# for run, sketch_rank in tqdm(list(product(runs, sketch_ranks)), desc="OTTS+3"):
#     # tensor = TensorTrain.random(shape, rank=tt_rank)
#     # tensor = tt_exp_decay(shape, tt_rank, seed=SEED + run)
#     tensor = tensors[run]
#     experiment.do_experiment(
#         tensor,
#         "OTTS+3",
#         experiment_orthogonal_sketch,
#         left_rank=sketch_rank,
#         right_rank=sketch_rank + 3,
#         tensor_rank=tt_rank,
#         sketch_rank=sketch_rank,
#         run=run,
#         error_func=tt_error_func,
#     )

# for run, sketch_rank in tqdm(list(product(runs, sketch_ranks)), desc="OTTSx2"):
#     # tensor = TensorTrain.random(shape, rank=tt_rank)
#     # tensor = tt_exp_decay(shape, tt_rank, seed=SEED + run)
#     tensor = tensors[run]
#     experiment.do_experiment(
#         tensor,
#         "OTTSx2",
#         experiment_orthogonal_sketch,
#         left_rank=sketch_rank,
#         right_rank=sketch_rank * 2,
#         tensor_rank=tt_rank,
#         sketch_rank=sketch_rank,
#         run=run,
#         error_func=tt_error_func,
#     )

for run, sketch_rank in tqdm(list(product(runs, sketch_ranks)), desc="HMT"):
    # tensor = TensorTrain.random(shape, rank=tt_rank, seed=SEED + run)
    # tensor = tt_exp_decay(shape, tt_rank)
    tensor = tensors[run]
    experiment.do_experiment(
        tensor,
        "HMT",
        experiment_hmt_sketch,
        rank=sketch_rank,
        tensor_rank=tt_rank,
        sketch_rank=sketch_rank,
        run=run,
        error_func=tt_error_func,
    )

# --------------------------------

# %%
import pandas as pd
import matplotlib.pyplot as plt

df = pd.read_csv(csv_filename)
label_dic = {
    "STTAx2": "STTAx2, TT-DRM",
    "STTA+3": "STTA+3, TT-DRM",
    "HMT": "TT-HMT, TT-DRM",
}
df = df[df["name"].isin(label_dic)]


def make_percentile_function(percentile):
    def f(x):
        return np.percentile(x, percentile)

    return f


percentile80 = make_percentile_function(80)
percentile20 = make_percentile_function(20)

timing_df = (
    df.groupby(["name", "sketch_rank", "tensor_rank"])[["time_taken", "error"]]
    .agg(
        min_time=("time_taken", "min"),
        time50=("time_taken", "median"),
        std_time=("time_taken", "std"),
        error50=("error", "median"),
        std_error=("error", "std"),
        time20=("time_taken", percentile20),
        time80=("time_taken", percentile80),
        error80=("error", percentile80),
        error20=("error", percentile20),
    )
    .reset_index()
)
max_xtick = np.ceil(timing_df.time50.max() * 10 + 1) / 10
timing_df.name.unique()
# %%
markers = ["s", "o", "D"]
for i, label in enumerate(("HMT", "STTA+3", "STTAx2")):
    sub_df = timing_df[(timing_df.name == label)]

    plt.plot(
        sub_df.time50,
        sub_df.error50,
        # xerr=[sub_df.time50 - sub_df.time20, sub_df.time80 - sub_df.time50],
        # yerr=[sub_df.error50 - sub_df.error20, sub_df.error80 - sub_df.error50],
        # capsize=3,
        marker=markers[i],
        ms=4,
        label=label_dic[label],
        # linestyle="",
    )

plt.ylabel("Relative error")
plt.xlabel("Time taken (s)")
plt.legend()
plt.yscale("log")
# plt.xscale("log")
plt.xlim(0, None)
plt.title("Error vs. time taken")
plt.xticks(np.arange(0, max_xtick, 0.1))
plt.savefig("results/timings1.pdf", transparent=True, bbox_inches="tight")

# %%
markers = ["s", "o", "D"]
for i, label in enumerate(("HMT", "STTA+3", "STTAx2")):
    sub_df = timing_df[(timing_df.name == label)]

    plt.plot(
        sub_df.sketch_rank,
        sub_df.time50,
        # yerr=[sub_df.time50 - sub_df.time20, sub_df.time80 - sub_df.time50],
        # capsize=3,
        label=label_dic[label],
        marker=markers[i],
        ms=4
        # linestyle="",
    )

plt.xlabel("TT-rank of sketch")
plt.ylabel("Time taken (s)")
plt.legend()
plt.yticks(np.arange(0, max_xtick, 0.2))
# plt.yscale("log")
plt.ylim(0, None)
plt.xlim(0, None)
# plt.xticks(np.concatenate([np.arange(5, 151, 10)]))
plt.xticks(timing_df.sketch_rank.unique())
plt.title("Time taken vs. sketch rank")
plt.savefig("results/timings2.pdf", transparent=True, bbox_inches="tight")
# %%
timing_df.sketch_rank.unique()



import urllib.request
import os.path
import os
from tqdm import tqdm

DATA_FOLDER = "data"


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(
        unit="B",
        unit_scale=True,
        miniters=1,
        desc="downloading " + url.split("/")[-1],
    ) as t:
        urllib.request.urlretrieve(
            url, filename=output_path, reporthook=t.update_to
        )


def download_data(url):
    try:
        os.mkdir(DATA_FOLDER)
    except FileExistsError:
        pass

    filename = url.split("/")[-1]
    filepath = os.path.join(DATA_FOLDER, filename)
    if os.path.isfile(filepath):
        print("File already downloaded")
    else:
        download_url(url, filepath)



# %%
"""Experiment of compressing a sum of TT tensors"""
import numpy as np
from typing import List
from itertools import product
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor

from experiment_base import (
    Experiment,
    experiment_orthogonal_sketch,
    experiment_stream_sketch,
    experiment_tt_svd,
    experiment_hmt_sketch,
)

n_dims = 5
dim = 10
shape = (dim,) * n_dims

tt_rank = 3
num_tts = 20

np.random.seed(179)
tt_summands_list: List[Tensor] = []
coeffs = np.logspace(0, -20, num_tts)
for coeff in coeffs:
    tensor_tt = TensorTrain.random(shape, rank=tt_rank) * coeff
    tt_summands_list.append(tensor_tt)
tensor = TensorSum(tt_summands_list)
tt_sum_dense = tensor.dense()
csv_filename = "results/tt_sum.csv"
experiment = Experiment(csv_filename)

# %%

ranks = range(1, 31)
runs = range(30)

for rank, run in tqdm(list(product(ranks, runs)), desc="OTTS"):
    experiment.do_experiment(
        tensor,
        "OTTS",
        experiment_orthogonal_sketch,
        left_rank=rank,
        right_rank=rank * 2,
        run=run,
    )

for rank, run in tqdm(list(product(ranks, runs)), desc="STTA"):
    experiment.do_experiment(
        tensor,
        "STTA",
        experiment_stream_sketch,
        left_rank=rank,
        right_rank=rank * 2,
        run=run,
    )

for rank, run in tqdm(list(product(ranks, runs)), desc="HMT"):
    experiment.do_experiment(
        tensor,
        "HMT",
        experiment_hmt_sketch,
        rank=rank,
        run=run,
    )

for rank in tqdm(ranks, desc="TT-SVD"):
    experiment.do_experiment(tensor, "TT-SVD", experiment_tt_svd, rank=rank)

# %%
import pandas as pd

df = pd.read_csv(csv_filename)

plt.figure(figsize=(10, 4))
ttsvd = df[df["name"] == "TT-SVD"]

# rsketch = df[df["name"] == "OTTS"]

# error_gb = rsketch.groupby(rsketch["left_rank"]).error
# errors05 = error_gb.quantile(0.5).values
# errors08 = error_gb.quantile(0.8).values - errors05
# errors02 = errors05 - error_gb.quantile(0.2).values
# plt.errorbar(
#     plot_ranks - 0.05,
#     errors05,
#     yerr=np.stack([errors02, errors08]),
#     label="OTTS, TT-DRM",
#     capsize=3,
#     linestyle="",
# )


hsketch = df[df["name"] == "HMT"]
plot_ranks = hsketch["rank"].unique()
plt.plot(plot_ranks, ttsvd.error.values, "-o", label="TT-SVD", ms=3)
error_gb = hsketch.groupby("rank").error
errors05 = error_gb.quantile(0.5).values
errors08 = error_gb.quantile(0.8).values - errors05
errors02 = errors05 - error_gb.quantile(0.2).values
plt.errorbar(
    plot_ranks - 0.1,
    errors05,
    yerr=np.stack([errors02, errors08]),
    label="TT-HMT, TT-DRM",
    capsize=3,
    linestyle="",
)

ssketch = df[df["name"] == "STTA"]
error_gb = ssketch.groupby("left_rank").error
errors05 = error_gb.quantile(0.5).values
errors08 = error_gb.quantile(0.8).values - errors05
errors02 = errors05 - error_gb.quantile(0.2).values
plt.errorbar(
    plot_ranks + 0.1,
    errors05,
    yerr=np.stack([errors02, errors08]),
    label="STTA, TT-DRM",
    capsize=3,
    linestyle="",
)
plt.xticks(ranks)
plt.ylabel("Relative error")
plt.xlabel("TT-rank")
plt.yscale("log")
plt.title(f"Approximation of sum of TT")
plt.legend()
plt.savefig("results/plot-tt-sum.pdf", transparent=True, bbox_inches="tight")
plt.show()

# %%



# cython: language_level=3
# cython: profile=True

import cython
import numpy as np
from libc.math cimport frexp, floor
cimport numpy as np
cimport scipy.special.cython_special
np.import_array()
ctypedef unsigned long long int uint64 
ctypedef unsigned int uint

@cython.boundscheck(False)
cpdef hash_int_c(uint64[:] vals):
    """Use simple hashing for generating random numbers from indices
    deterministically. See
    https://stackoverflow.com/questions/664014/what-integer-hash-function-are-good-that-accepts-an-integer-hash-key
    """
    cdef Py_ssize_t N = vals.shape[0]
    cdef uint shift1 = 30
    cdef uint shift2 = 27
    cdef uint shift3 = 31
    cdef uint64 mult1 = 0xBF58476D1CE4E5B9
    cdef uint64 mult2 = 0x94D049BB133111EB
    cdef uint64 add1 = 0X4BE98134A5976FD3

    cdef int i
    cdef uint64 r
    for i in range(N):
        r = vals[i]
        r = r + add1
        r ^= r >> shift1
        r *= mult1
        r ^= r >> shift2
        r *= mult2
        r ^= r >> shift3
        vals[i] = r

@cython.boundscheck(False)
cdef uint_to_normal(double[:] nums):
    """Turn hashed indices into normally distributed numbers"""
    # cdef double[:] nums = <double[:]>inds
    cdef Py_ssize_t N = nums.shape[0]
    cdef Py_ssize_t i
    cdef int e
    cdef double x
    for i in range(N): #prange here messes stuff up when parallelizing
        x = frexp(nums[i], &e) * 2 - 1
        x = scipy.special.cython_special.ndtri(x)
        nums[i] = x

@cython.boundscheck(False)
cpdef double[:] _inds_to_rand_double(
    uint64[:,:] indices, 
    uint64[:] shape,
    int rank_min,
    int rank_max,
    uint64 seed,
):  
    cdef uint64[:] indices_flat = np.copy(indices[0])
    cdef int N = indices.shape[1]
    cdef int n_dim = len(shape)
    cdef int rank = rank_max - rank_min

    cdef int prod = shape[0]
    cdef int i
    cdef int j
    for i in range(1,n_dim):
        for j in range(N):
            indices_flat[j] += indices[i,j] * prod
        prod *= shape[i]
    
    cdef uint64[:] size_salt = np.arange(rank_min, rank_max, dtype=np.uint64)
    hash_int_c(size_salt)
    # print(np.array(size_salt)) (if we increase rank then first r size salts
    # are still the same)
    
    for i in range(rank):
        size_salt[i] += seed

    cdef uint64[:,:] indices_flat_large = np.empty(
        (N, rank), dtype=np.uint64
    )
    
    for i in range(N):
        for j in range(rank):
            indices_flat_large[i,j] = indices_flat[i] + size_salt[j]
    
    cdef uint64[:] hashes =  np.reshape(indices_flat_large, -1)

    cdef uint64 SETBIT3 = 0x2000000000000000
    cdef uint64 SETBITS12 = 0x3FFFFFFFFFFFFFFF
    cdef uint64 h
    hash_int_c(hashes)
    for i in range(N*rank):
        # Set first three bits to 001
        # Avoids NaN values and zero values for float exp
        # Assures mantissa is between 0.5 and 1.0 
        h = hashes[i]
        h = (h | SETBIT3) & SETBITS12 
        hashes[i] = h
    cdef double[:] nums = np.frombuffer(hashes, dtype=np.float64)
    

    return nums


cdef double[:] _inds_to_normal(
    uint64[:,:] indices, 
    uint64[:] shape,
    int rank_min,
    int rank_max,
    uint64 seed,
):  
    cdef double[:] nums = _inds_to_rand_double(
        indices, shape, rank_min, rank_max, seed
    )
    uint_to_normal(nums)
    return nums

@cython.boundscheck(False)
cdef short[:] _inds_to_sparse_sign(
    uint64[:,:] indices, 
    uint64[:] shape,
    int rank,
    int nnz,
    uint64 seed,
):
    cdef Py_ssize_t N = indices.shape[1]
    cdef double[:] uniform_nums = _inds_to_rand_double(
        indices, shape, 0, nnz, seed
    )
    cdef short[:] sparse_sign = np.zeros((N*rank), dtype=np.int16)
    cdef int i
    cdef int j
    cdef int e
    cdef short temp
    cdef int rand_num

    for i in range(N):
        # populate matrix with rand signs
        for j in range(nnz):
            # Replace uniform num with mantissa, store exponent in &e
            uniform_nums[i*nnz+j] = frexp(uniform_nums[i*nnz+j], &e) * 2 - 1
            sparse_sign[i*rank+j] = (e % 2) * 2 - 1  # random sign
        
        # randomly permute
        for j in range(nnz):
            # random int in [j,rank)
            rand_num = <int>(uniform_nums[i*nnz+j]*(rank-j) + j) 
            temp = sparse_sign[i*rank+j]
            sparse_sign[i*rank+j] = sparse_sign[i*rank+rand_num]
            sparse_sign[i*rank+rand_num]=temp
    return sparse_sign

def inds_to_sparse_sign(
    indices,
    shape,
    rank,
    rank_min,
    rank_max,
    non_zero_per_row,
    seed
):
    """
    Converts a list of indices into the non-zero entries of a sparse sign matrix
    """
    indices = indices.astype(np.uint64)
    N = indices.shape[1]
    shape = np.array(shape,dtype=np.uint64)
    rank_min = int(rank_min)
    rank_max = int(rank_max)
    non_zero_per_row = int(non_zero_per_row)
    seed = np.mod(seed,2**63,dtype=np.uint64)
    nums_view = _inds_to_sparse_sign(
        indices, shape, rank, non_zero_per_row, seed
    )
    nums = np.array(nums_view).reshape(N,rank)
    nums = nums[:, rank_min:rank_max]
    return nums


def inds_to_normal(
    indices,
    shape,
    rank_min,
    rank_max,
    seed
):  
    """
    Converts a list of indices into the associated entries of gaussian matrix
    """
    indices = indices.astype(np.uint64)
    N = indices.shape[1]
    shape = np.array(shape,dtype=np.uint64)
    rank_min = int(rank_min)
    rank_max = int(rank_max)
    rank = rank_max - rank_min
    seed = np.mod(seed,2**63,dtype=np.uint64)
    nums_view = _inds_to_normal(indices, shape, rank_min, rank_max, seed)
    nums = np.array(nums_view)
    return nums.reshape(N,rank)


# %%
"""In this experiment we see how much the error can improve when we use TT-SVD
rounding to a fixed rank after first oversampling. Benchmark is then the TT-SVD error."""


from pydoc import describe
import re
import numpy as np
from typing import List
from itertools import product
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor

from experiment_base import (
    Experiment,
    experiment_stream_sketch,
    experiment_tt_svd,
    experiment_hmt_sketch,
)

n_dims = 5
dim = 10
shape = (dim,) * n_dims

tt_rank = 3
num_tts = 20

np.random.seed(179)
tt_summands_list: List[Tensor] = []
coeffs = np.logspace(0, -20, num_tts)
for coeff in coeffs:
    tensor_tt = TensorTrain.random(shape, rank=tt_rank) * coeff
    tt_summands_list.append(tensor_tt)
tensor = TensorSum(tt_summands_list)
tt_sum_dense = tensor.dense()
csv_filename = "results/recompression.csv"
experiment = Experiment(csv_filename)

# %%

recompression_rank = 10
sketch_ranks = range(recompression_rank, recompression_rank + 20)
runs = range(30)

for rank, run in tqdm(list(product(sketch_ranks, runs)), desc="HMT"):
    experiment.do_experiment(
        tensor,
        "HMT",
        experiment_hmt_sketch,
        rank=rank,
        recompression_rank=recompression_rank,
        run=run,
    )

for rank, run in tqdm(list(product(sketch_ranks, runs)), desc="STTA+3"):
    experiment.do_experiment(
        tensor,
        "STTA+3",
        experiment_stream_sketch,
        left_rank=rank,
        right_rank=rank + 3,
        recompression_rank=recompression_rank,
        run=run,
    )

for rank, run in tqdm(list(product(sketch_ranks, runs)), desc="STTAx2"):
    experiment.do_experiment(
        tensor,
        "STTAx2",
        experiment_stream_sketch,
        left_rank=rank,
        right_rank=rank + 3,
        recompression_rank=recompression_rank,
        run=run,
    )
experiment.do_experiment(
    tensor, "tt_svd", experiment_tt_svd, rank=recompression_rank
)

# %%
import pandas as pd
plt.figure(figsize=(8, 4))

df = pd.read_csv(csv_filename, index_col=0)
tt_svd_erorr = df[df["name"]=="tt_svd"].iloc[0]["error"]
# plt.axhline(tt_svd_erorr, ls="--", color="k", label="TT-SVD")

label = "HMT"
rank_label = "rank"
plot_pairs = [("HMT", "rank"), ("STTA+3", "left_rank"), ("STTAx2", "left_rank")]
markers = ["s","v","^"]
for i, (label, rank_label) in enumerate(plot_pairs):
    df_method = df[df["name"] ==label].copy()
    df_method["log_error"] = np.log10(df_method["error"] / tt_svd_erorr)
    ranks = df_method[rank_label].unique()
    error80 = df_method.groupby(rank_label)["log_error"].quantile(0.8)
    error50 = df_method.groupby(rank_label)["log_error"].median()
    error20 = df_method.groupby(rank_label)["log_error"].quantile(0.2)

    jitter = np.linspace(-0.2, 0.2, len(plot_pairs))[i]
    plt.errorbar(
        ranks+jitter,
        error50,
        yerr=[error50 - error20, error80 - error50],
        capsize=2,
        label=label,
        linestyle="",
        marker=markers[i],
        ms="3"
    )
plt.xticks(ranks)
plt.yscale("log")
plt.xlabel("TT-rank")
plt.ylabel("Error relative to TT-SVD")
plt.legend()

locs_new = np.array([10,3,1.8,1.33,1.15,1.075,1.04,1.02,1.01,1.005])
labels_new =[str(s) for s in locs_new]
locs_new = np.log10(locs_new)
plt.yticks(locs_new, labels_new)
plt.minorticks_off()

# %%
locs_new
10**np.logspace(np.log10(1),np.log10(0.002), 10)


r"""Implements sketching methods for given types of tensors.

Each sketching method consists of a method to sketch ``omega``
:math:`\Omega_\mu` and to sketch ``psi`` :math:`\Psi_\mu` (following the
notation of the paper), using the output from the respective ``DRM.sketch_``
methods. 

For example, the ``sketch_omega_tt`` method sketches the tensor
:math:`\Omega_\mu` using the output of ``some_drm.sketch_tt``, where
``some_drm`` is any DRM that supports the ``CanSketchTT`` interface.
"""


# %%
import logging
from itertools import product
from time import perf_counter
from typing import Tuple

import numpy as np
from tqdm import tqdm
from tt_sketch.tensor import TensorSum, TensorTrain
from tt_sketch.tt_gmres import ROUNDING_MODE, round_tt_sum, tt_sum_gmres

from cookie_problem import prepare_cookie_problem
from experiment_base import Experiment

LOG_FILE = "cookie.log"
logging.basicConfig(
    filename=LOG_FILE,
    level=logging.INFO,
    format="%(asctime)s: %(message)s",
)

map_sum, B, precond_map = prepare_cookie_problem(100, 4)
x0 = TensorTrain.zero(shape=map_sum.in_shape, rank=1)

csv_filename = "results/cookie_2x2.csv"
experiment = Experiment(csv_filename)


def experiment_cookie(
    _,
    rounding_method: ROUNDING_MODE,
    rounding_method_final: ROUNDING_MODE,
    max_rank: int,
    final_round_ranks: Tuple[int, ...],
    **kwargs,
):
    logging.info("\n")
    logging.info("-" * 80)
    logging.info(f"Rounding method: {rounding_method}")
    logging.info(f"Rounding method final: {rounding_method_final}")
    logging.info(f"Max rank: {max_rank}")
    logging.info(f"Final round ranks: {final_round_ranks}")
    logging.info(f"Kwargs: {kwargs}")
    logging.info("\n")
    result, history = tt_sum_gmres(
        A=map_sum,
        b=B,
        precond=precond_map,
        x0=x0,
        max_rank=max_rank,
        final_round_rank=None,
        maxiter=50,
        tolerance=0,
        symmetric=True,
        rounding_method=rounding_method,
        rounding_method_final=None,
        verbose=True,
    )
    experiment_result = {}
    experiment_result["last_residual"] = history["residual_norm"][-1]
    experiment_result["gmres_time"] = np.sum(history["step_time"])
    # experiment_result["final_round_time"] = history["final_round_time"]

    true_errors = []
    precond_errors = []
    round_timings = []
    for rank in final_round_ranks:
        current_time = perf_counter()
        result_rounded = round_tt_sum(
            result, rank, method=rounding_method_final, eps=None  # type: ignore
        )
        round_timings.append(perf_counter() - current_time)
        true_error = map_sum(result_rounded).error(B, relative=True)
        precond_error = TensorSum(
            [precond_map(t) for t in map_sum(result_rounded).tensors]
        ).error(precond_map(B), relative=True)
        true_errors.append(true_error)
        precond_errors.append(precond_error)
        logging.info(f"Rank {rank}: true error: {true_error:.4e}")
        logging.info(f"Rank {rank}: preconditioned error: {precond_error:.4e}")

    experiment_result["true_error"] = np.array(true_errors)
    experiment_result["precond_error"] = np.array(precond_errors)
    experiment_result["final_round_time"] = np.array(round_timings)

    return experiment_result


sketch_ranks = list(range(10, 101, 5))
pairwise_ranks = list(range(10, 51, 5))
round_methods = ["sketch", "pairwise"]
final_round_methods = ["sketch", "pairwise"]
final_round_ranks = tuple(range(10, 101, 5))
runs = list(range(20))


all_experiments = []
for run in runs:
    all_experiments.extend(
        list(
            product(["sketch"], ["sketch"], sketch_ranks, [final_round_ranks], [run])
        )
    )
    all_experiments.extend(
        list(
            product(
                ["pairwise"], ["sketch"], pairwise_ranks, [final_round_ranks], [run]
            )
        )
    )
    all_experiments.extend(
        list(
            product(
                ["sketch"], ["pairwise"], sketch_ranks, [final_round_ranks], [run]
            )
        )
    )
    all_experiments.extend(
        list(
            product(
                ["pairwise"],
                ["pairwise"],
                pairwise_ranks,
                [final_round_ranks],
                [run],
            )
        )
    )

for (
    rounding_method,
    rounding_method_final,
    max_rank,
    final_round_ranks,
    run,
) in tqdm(all_experiments):
    experiment.do_experiment(
        None,
        "cookie_2x2",
        experiment_cookie,
        rounding_method=rounding_method,
        rounding_method_final=rounding_method_final,
        max_rank=max_rank,
        final_round_ranks=final_round_ranks,
        run=run,
    )



# %%
"""
In this script we investigate the relationship between error and timing for the
three different method. We want to see if OTTS makes sense versus the HMT
method. Despite the fact that OTTS is better, it's also a little slower. We need
to see if there are situations where OTTS can lead to a better approximation in
the same amount of time as HMT.
"""

from unittest import result
import numpy as np
import pandas as pd
from typing import List
from experiment_base import (
    Experiment,
    experiment_orthogonal_sketch,
    experiment_stream_sketch,
    experiment_hmt_sketch,
    experiment_tt_svd,
)
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.sketch import stream_sketch, orthogonal_sketch, hmt_sketch
from itertools import product
import time
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor, SparseTensor

n_dims = 5
dim = 1000
shape = (dim,) * n_dims

tt_rank = 100
num_tts = 10

np.random.seed(179)
tt_summands_list: List[Tensor] = []
coeffs = np.logspace(0, -10, num_tts)
for coeff in coeffs:
    tensor_tt = TensorTrain.random(shape, rank=tt_rank) * coeff
    tt_summands_list.append(tensor_tt)
tensor = TensorSum(tt_summands_list)
# tt_sum_dense = tensor.dense()

# n_dims = 6
# dim = 5
# shape = (dim,) * n_dims

# tt_rank = 3

# np.random.seed(179)
# tensor_tt = TensorTrain.random(shape, rank=tt_rank)

# nnz = 100
# tot_dim = np.prod(shape)
# inds_dense = np.random.choice(tot_dim, nnz, replace=False)
# inds = np.stack(np.unravel_index(inds_dense, shape))
# entries = np.random.normal(size=nnz)
# entries = entries * np.logspace(-3, -20, nnz)
# tensor_sparse = SparseTensor(shape, inds, entries)

# tensor = tensor_sparse + tensor_tt


# %%
import cProfile
cProfile.run('hmt_sketch(tensor, rank=10)')
# %%
max_rank = 5
results = []
runs = range(10)
ranks = range(max_rank, max_rank + 20)
# ranks = [200]


def hmt_func(tensor, left_rank, right_rank):
    return hmt_sketch(tensor, rank=left_rank)


method_dict = {
    "OTTS": orthogonal_sketch,
    "STTA": stream_sketch,
    "HMT": hmt_func,
}
for run, r, method in tqdm(list(product(runs, ranks, method_dict.keys()))):
    current_time = time.perf_counter()
    stt = method_dict[method](tensor, left_rank=r, right_rank=r+1)
    if not isinstance(stt, TensorTrain):
        tt = stt.to_tt()
    else:
        tt = stt
    orthogonalize = method != "STTA"
    # tt = tt.round(max_rank=max_rank, orthogonalize=orthogonalize)
    time_taken = time.perf_counter() - current_time
    error = tt.error(tensor, fast=False, relative=True)
    results.append(
        {
            "method": method,
            "rank": r,
            "time": time_taken,
            "error": error,
            "run": run,
        }
    )
# %%
shape_string = "x".join(str(x) for x in shape)
ranks_string = str(min(ranks)) + "-" + str(max(ranks))
plt.figure(figsize=(7, 5))
title = f"shape={shape_string}, {tt_rank=}, {num_tts=}, ranks={ranks_string}"
plt.title(title)
results_df = pd.DataFrame(results)
plot_df = (
    results_df.groupby(["method", "rank"])[["error", "time"]]
    .median()
    .reset_index()
)
for method in method_dict.keys():
    plt.plot(
        plot_df.loc[plot_df["method"] == method, "time"],
        plot_df.loc[plot_df["method"] == method, "error"],
        "o",
        label=method,
    )
plt.legend()
plt.yscale("log")
plt.xlabel("Time (s)")
plt.ylabel("Error")

# %%



from abc import ABC, abstractmethod
from typing import Tuple

from tt_sketch.drm_base import DRM
from tt_sketch.tensor import (
    CPTensor,
    DenseTensor,
    SparseTensor,
    TensorTrain,
    TuckerTensor,
)
from tt_sketch.utils import ArrayGenerator


class CansketchTT(DRM, ABC):
    @abstractmethod
    def sketch_tt(self, tensor: TensorTrain) -> ArrayGenerator:
        r"""List of contractions of form :math:`Y_\mu^\top\mathcal{T}^{\leq\mu}` where
        :math:`Y_\mu` is the DRM, and :math:`\mathcal T^{\leq\mu}` the
        contraction of the first :math:`\mu` cores of ``tensor``.

        Returns array of shape ``(tensor.rank[mu], drm.rank[mu])``"""


class CansketchSparse(DRM, ABC):
    rank: Tuple[int, ...]

    @abstractmethod
    def sketch_sparse(self, tensor: SparseTensor) -> ArrayGenerator:
        """Computes list of sketching matrices sampled into a vector using the
        indices of ``tensor`` for each unfolding. Shape of each vector is
        ``v[mu] = (rank[mu], tensor.nnz)``. This way the contraction between
        ``tensor`` and the sketching matrix is of form ``np.dot(tensor.entries,
        v[mu])``"""


class CansketchDense(DRM, ABC):
    @abstractmethod
    def sketch_dense(self, tensor: DenseTensor) -> ArrayGenerator:
        r"""Return list of dense DRMs. Of shape
        ``(np.prod(tensor.shape[ :mu+1]), rank[mu])``"""


class CansketchCP(DRM, ABC):
    @abstractmethod
    def sketch_cp(self, tensor: CPTensor) -> ArrayGenerator:
        r"""List of contractions of form :math:`Y_\mu^\top\mathcal{T}^{\leq\mu}` where
        :math:`Y_\mu` is the DRM, and :math:`\mathcal T^{\leq\mu}` the
        contraction of the first :math:`\mu` cores of ``tensor``.

        Returns array of shape ``(tensor.rank[mu], drm.rank[mu])``"""


class CanSketchTucker(DRM, ABC):
    @abstractmethod
    def sketch_tucker(self, tensor: TuckerTensor) -> ArrayGenerator:
        r"""
        List of contractions of form :math:`Y_\mu^\top(U_1\otimes\cdots\otimes U_\mu)`
        where ``Y_\mu`` is the DRM, and :math:`U_\mu` denotes the factors of the
        Tucker decomposition.

        Returns array of shape ``(np.prod(tensor.rank[:mu]), drm.rank[mu])``
        """



from typing import Optional

import numpy as np
import numpy.typing as npt
from tt_sketch.tensor import TensorTrain


def sketch_omega_tt(
    left_sketch: npt.NDArray, right_sketch: npt.NDArray, **kwargs
):
    return left_sketch.T @ right_sketch


def sketch_psi_tt(
    left_sketch: Optional[npt.NDArray],
    right_sketch: Optional[npt.NDArray],
    *,
    tensor: TensorTrain,
    mu: int,
    **kwargs
):
    tt_core = tensor.cores[mu]
    if left_sketch is None:
        Psi = np.einsum("ijk,kl->ijl", tt_core, right_sketch)  # type: ignore
    elif right_sketch is None:
        Psi = np.einsum("ij,jkl->ikl", left_sketch.T, tt_core)
    else:
        Psi = np.einsum(
            "ij,jkl,lm->ikm",
            left_sketch.T,
            tt_core,
            right_sketch,
            optimize="optimal",
        )
    return Psi



# %%
"""Compress high-rank CP obtained from random forest to TT

We can write a decision tree as a low-rank CP. Since random forests are just
weighted sums of decision trees, they too can be expressed as a CP. To limit the
dimension of the resulting tensor we do need to compress the dimension of this
CP by taking a subset of the thresholds.
"""

from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from ttml.forest_compression import compress_forest_thresholds, forest_to_CP
import numpy as np
from tt_sketch.tensor import CPTensor
from tt_sketch.sketch import stream_sketch
from tqdm import tqdm
import matplotlib.pyplot as plt
import pandas as pd
from data_download_util import download_data

FILE_URL = (
    "https://archive.ics.uci.edu/ml/machine-learning-databases/00291"
    "/airfoil_self_noise.dat"
)
download_data(FILE_URL)
df = pd.read_csv("data/airfoil_self_noise.dat", sep="\t", header=None)
Xy = df.to_numpy()
print(Xy.shape)

scaler = StandardScaler()
Xy = scaler.fit_transform(Xy)

y = Xy[:, 5]
X = Xy[:, :5]
X_train, X_val, y_train, y_val = train_test_split(
    X,
    y,
    test_size=0.2,
    random_state=179,
)
forest = RandomForestRegressor(
    max_leaf_nodes=500, max_depth=None, n_estimators=50
)
forest.fit(X_train, y_train)


def mse(estim):
    return np.mean((estim.predict(X_val) - y_val) ** 2)


uncompressed_loss = mse(forest)
uncompressed_loss

# %%
num_thresholds = np.arange(25, 100, 5)
errors = []
for num_t in tqdm(num_thresholds):
    c_forest, _ = compress_forest_thresholds(forest, num_t)
    errors.append(mse(c_forest))

plt.plot(num_thresholds, errors, ".")
plt.axhline(uncompressed_loss, ls="--", c="k")

# %%
"""40 seems like a good number of thresholds for this forest"""
c_forest, thresh = compress_forest_thresholds(forest, 10)
cp_cores = forest_to_CP(c_forest, thresh, take_mean=True)
cp_tensor = CPTensor(cp_cores)

N = 1000
dim = np.prod(cp_tensor.shape)
dense_idx = np.random.choice(dim, size=N)
idx = np.unravel_index(dense_idx, cp_tensor.shape)
cp_samples = cp_tensor.gather(idx)

ranks = range(5, 50, 1)
errors = []
for rank in tqdm(ranks):
    stt = stream_sketch(cp_tensor, left_rank=rank, right_rank=rank*2)
    tt = stt.to_tt()
    error = np.linalg.norm(cp_samples - tt.gather(idx)) / np.linalg.norm(
        cp_samples
    )
    errors.append(error)

# %%

plt.plot(ranks, errors)
plt.yscale("log")

"""Seems like it works, but the errors are just way too high. That kinda shows
that this can't be compressed well as a tt, which should not come as a
surprise..."""
# %%



"""Solve the cookie problem using TT-GMRES. 

The files ``cookies_matrices_2x2.mat`` and ``cookies_matrices_3x3.mat`` can be obtained from the 'examples' folder of the ``htucker`` package.
https://www.epfl.ch/labs/anchp/index-html/software/htucker/
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Dict, Generic, List, Optional, Tuple, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.io
import scipy.linalg
from time import perf_counter
from tqdm import tqdm
from tt_sketch.sketch import stream_sketch
from tt_sketch.tensor import TensorSum, TensorTrain
from tt_sketch.tt_gmres import (
    TTLinearMap,
    TTPrecond,
    TTLinearMapSum,
)
from tt_sketch.utils import (
    TTRank,
)


COOKIES_2x2_FILE = "data/cookies_matrices_2x2.mat"
COOKIES_3x3_FILE = "data/cookies_matrices_3x3.mat"


class CookieMap(TTLinearMap):
    def __init__(
        self,
        A: npt.NDArray,
        mode: int,
        shape: Tuple[int, ...],
        coeffs: npt.NDArray,
    ) -> None:
        self.A = A
        self.mode = mode
        self.in_shape = shape
        self.out_shape = shape
        self.coeffs = coeffs

    def cookie_call(self, other: TensorTrain) -> TensorTrain:
        new_cores = deepcopy(other.cores)
        new_cores[0] = np.einsum("ijk,jl->ilk", new_cores[0], self.A)
        if self.mode != 0:
            new_cores[self.mode] = np.einsum(
                "ijk,j->ijk", new_cores[self.mode], self.coeffs
            )
        tt = TensorTrain(new_cores)
        return tt

    __call__ = cookie_call


def prepare_cookie_problem(num_coeffs, num_cookies):
    if num_cookies == 4:
        cookies_file = COOKIES_2x2_FILE
    elif num_cookies == 9:
        cookies_file = COOKIES_3x3_FILE
    else:
        raise ValueError("num_cookies must be 4 or 9")
    cookies_data = scipy.io.loadmat(cookies_file)
    b = cookies_data["b"].reshape(-1)
    A_list = cookies_data["A"][0]

    shape = (A_list[0].shape[0],) + (num_coeffs,) * (len(A_list) - 1)

    A_precond_list = []
    coeffs_list = []
    for mu, A in enumerate(A_list):
        A = A.toarray()
        if mu == 0:
            coeffs = np.ones(A.shape[0])
        else:
            coeffs = np.linspace(0, 10, num_coeffs, dtype=np.float64)
        A_precond_list.append(A * np.mean(coeffs))
        coeffs_list.append(coeffs)

    A_precond = np.sum(A_precond_list, axis=0)
    precond_map = TTPrecond(A_precond, shape, mode=0)

    cookie_maps = []

    for mu, (A, coeffs) in enumerate(zip(A_list, coeffs_list)):
        cookie_maps.append(
            CookieMap(
                A.toarray(),
                mu,
                shape,
                coeffs,
            )
        )

    map_sum = TTLinearMapSum(cookie_maps)

    B_cores = [b.reshape(1, -1, 1)]
    for n in shape[1:]:
        B_cores.append(np.ones((1, n, 1)))
    B = TensorTrain(B_cores)

    return map_sum, B, precond_map




from typing import Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from tt_sketch.tensor import SparseTensor


def _Psi_core_slice(
    inds: npt.NDArray[Union[np.int64, np.uint64]],
    entries: npt.NDArray,
    left_sketch_vec: Optional[npt.NDArray],
    right_sketch_vec: Optional[npt.NDArray],
    mu: int,
    j: int,
) -> npt.NDArray[np.float64]:
    """Compute slice of tensor Y[:,j,:] from left+right sketching matrix
    and entries+index of sparse tensor"""
    mask = inds[mu] == j

    if mu == 0:  # only sketch on right
        if right_sketch_vec is None:
            raise ValueError
        Psi_slice = entries[mask] @ right_sketch_vec[:, mask].T
        Psi_slice = Psi_slice.reshape(1, -1)
    elif mu == inds.shape[0] - 1:  # only sketch on left
        if left_sketch_vec is None:
            raise ValueError
        Psi_slice = left_sketch_vec[:, mask] @ entries[mask]
        Psi_slice = Psi_slice.reshape(-1, 1)
    else:  # sketch on both sides
        if left_sketch_vec is None or right_sketch_vec is None:
            raise ValueError
        Psi_slice = (
            left_sketch_vec[:, mask] * entries[mask]
        ) @ right_sketch_vec[:, mask].T
    return Psi_slice


def sketch_omega_sparse(
    left_sketch: npt.NDArray,
    right_sketch: npt.NDArray,
    *,
    tensor: SparseTensor,
    **kwargs,
) -> npt.NDArray:
    return (left_sketch * tensor.entries) @ right_sketch.T


def sketch_psi_sparse(
    left_sketch: npt.NDArray,
    right_sketch: npt.NDArray,
    *,
    tensor: SparseTensor,
    mu: int,
    psi_shape: Tuple[int, int, int],
    **kwargs,
) -> npt.NDArray:
    Psi = np.zeros(psi_shape)
    n_mu = psi_shape[1]
    for j in range(n_mu):
        Psi[:, j, :] = _Psi_core_slice(
            tensor.indices,
            tensor.entries,
            left_sketch,
            right_sketch,
            mu,
            j,
        )
    return Psi



# %%
import concurrent.futures
import multiprocessing
from functools import reduce
from operator import mul
from typing import Generator, List, Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt
import scipy.linalg

# from numpy.random import Generator, SeedSequence, default_rng
from numpy.typing import ArrayLike

ArrayList = List[npt.NDArray[np.float64]]
ArrayGenerator = Generator[npt.NDArray[np.float64], None, None]
TTRank = Union[int, Tuple[int, ...]]


def hilbert_tensor(n_dims: int, size: int) -> npt.NDArray:
    """Create a Hilbert tensor of specified size and dimensionality."""
    grid = np.meshgrid(*([np.arange(size)] * n_dims))
    hilbert = 1 / (np.sum(np.array(grid), axis=0) + 1)
    return hilbert


def sqrt_tensor(shape: Tuple[int, ...], a=-0.2, b=2) -> npt.NDArray:
    """Create a tensor of specified shape with square root of a sum a grid.

    Values of grid entries vary between a and b."""

    def sqrt_sum(X):
        return np.sqrt(np.abs(np.sum(X, axis=0)))

    vals = [np.linspace(a, b, s) for s in shape]
    grid = np.stack(np.meshgrid(*vals))
    X = sqrt_sum(grid)
    X /= np.linalg.norm(X)
    return X


def power_decay_tensor(
    shape: Tuple[int], pow: float = 2.0, seed=None
) -> ArrayLike:
    """Create tensor of specified shape such that singular values of each
    unfolding decay with a power law."""
    # if seed is not None:
    #     np.random.seed(np.mod(seed, 2**32 - 1))
    seq = SeedSequence(seed)
    A_seed = seq.generate_state(1)[0]
    A = random_normal(shape=shape, seed=A_seed)
    for mode in range(len(A.shape)):
        A_mat = matricize(A, mode)
        U, S, V = np.linalg.svd(A_mat, full_matrices=False)

        S /= S[0]
        S *= 1 / np.arange(1, len(S) + 1) ** pow
        A_mat = U @ np.diag(S) @ V
        A = dematricize(A_mat, mode, A.shape)
    return A


def matricize(
    A: npt.NDArray, mode: Union[int, Sequence[int]], mat_shape: bool = False
):
    """Matricize tensor ``A`` with respect to ``mode``.

    If mode is an int, return matrix. If mode is a tuple, return tensor of order
    ``len(mode)+1``, unless  ``mat_shape=True``"""
    if isinstance(mode, int):
        mode = (mode,)
    else:  # Try casting to tuple
        mode = tuple(mode)
    perm = mode + tuple(i for i in range(len(A.shape)) if i not in mode)
    A = np.transpose(A, perm)
    right_shape = (np.prod(A.shape[len(mode) :], dtype=int),)
    if mat_shape:
        left_shape = (np.prod(A.shape[: len(mode)], dtype=int),)
    else:
        left_shape = A.shape[: len(mode)]  # type: ignore
    A = A.reshape(left_shape + right_shape)

    return A


def dematricize(A, mode, shape):
    """Undo matricization of ``A`` with respect to ``mode``. Needs ``shape`` of
    original tensor."""
    current_shape = [A.shape[0]] + [s for i, s in enumerate(shape) if i != mode]
    current_shape = tuple(current_shape)
    A = A.reshape(current_shape)
    perm = list(range(1, len(shape)))
    perm = perm[:mode] + [0] + perm[mode:]
    A = np.transpose(A, perm)
    return A


def right_mul_pinv(A, B, cond=None):
    """Compute numerically stable product ``A@np.linalg.pinv(B)``"""
    lstsq = scipy.linalg.lstsq(B.T, A.T, cond=cond)

    return lstsq[0].T


def left_mul_pinv(A, B, cond=None):
    """Compute numerically stable product ``np.linalg.pinv(A)@B``"""
    lstsq = scipy.linalg.lstsq(A, B, cond=cond)

    return lstsq[0]


def projector(X: npt.NDArray, Y: Optional[npt.NDArray] = None) -> npt.NDArray:
    r"""Compute oblique projector :math:`\mathcal P_{X,Y}`"""
    if Y is None:
        Y = X

    P = X @ np.linalg.pinv(Y.T @ X) @ Y.T
    return P


def trim_ranks(
    dims: Tuple[int, ...], ranks: Tuple[int, ...]
) -> Tuple[int, ...]:
    """Return TT-rank to which TT can be exactly reduced

    A tt-rank can never be more than the product of the dimensions on the left
    or right of the rank. Furthermore, any internal edge in the TT cannot have
    rank higher than the product of any two connected supercores. Ranks are
    iteratively reduced  for each edge to satisfy these two requirements until
    the requirements are all satisfied.
    """
    ranks_trimmed = list(ranks)

    for i, r in enumerate(ranks_trimmed):
        dim_left = reduce(mul, dims[: i + 1], 1)
        dim_right = reduce(mul, dims[i + 1 :], 1)
        ranks_trimmed[i] = min(r, dim_left, dim_right)
    changed = True
    ranks_trimmed = [1] + ranks_trimmed + [1]
    for _ in range(100):
        changed = False
        for i, d in enumerate(dims):
            if ranks_trimmed[i + 1] > ranks_trimmed[i] * d:
                changed = True
                ranks_trimmed[i + 1] = ranks_trimmed[i] * d
            if ranks_trimmed[i] > d * ranks_trimmed[i + 1]:
                changed = True
                ranks_trimmed[i] = d * ranks_trimmed[i + 1]
        if not changed:
            break

    return tuple(ranks_trimmed[1:-1])


def process_tt_rank(
    rank: TTRank, shape: Tuple[int, ...], trim: bool
) -> Tuple[int, ...]:
    """
    Process TT rank, and check validity. Makes sure rank is a tuple.

    If ``trim=True``, ranks are trimmed to the smallest possible lossless value.
    """
    # check if rank is iterable, if not use constant rank
    try:
        rank_tuple = tuple(rank)  # type: ignore
    except TypeError:
        rank_tuple = (rank,) * (len(shape) - 1)  # type: ignore
    if len(rank_tuple) != len(shape) - 1:
        raise ValueError(
            f"TT-rank {rank_tuple} doesn't have right number of elements"
        )
    if trim:
        rank_tuple = trim_ranks(shape, rank_tuple)

    return rank_tuple


class MultithreadedRNG:
    """
    Multithreaded standard normal random number generator.

    Copy pasta from numpy docs
    """

    def __init__(self, shape, seed=None, threads=None):
        if threads is None:
            threads = multiprocessing.cpu_count()
        self.threads = threads

        seq = np.random.SeedSequence(seed)
        self._random_generators = [
            np.random.default_rng(s) for s in seq.spawn(threads)
        ]

        self.shape = shape
        n = np.prod(shape)
        self.executor = concurrent.futures.ThreadPoolExecutor(threads)
        self.values = np.empty(n)
        self.step = np.ceil(n / threads).astype(np.int_)
        self.fill()

    def fill(self):
        def _fill(random_state, out, first, last):
            random_state.standard_normal(out=out[first:last])

        futures = {}
        for i in range(self.threads):
            args = (
                _fill,
                self._random_generators[i],
                self.values,
                i * self.step,
                (i + 1) * self.step,
            )
            futures[self.executor.submit(*args)] = i
        concurrent.futures.wait(futures)
        self.values = self.values.reshape(self.shape)

    # def __del__(self):
    #     self.executor.shutdown(False)


def random_normal(shape, seed=None):
    """
    Generate multi-threaded random numbers
    """
    return MultithreadedRNG(shape, seed).values



import numpy as np
import numpy.typing as npt
from tt_sketch.tensor import CPTensor


def sketch_omega_cp(
    left_sketch: npt.NDArray, right_sketch: npt.NDArray, **kwargs
) -> npt.NDArray:
    return left_sketch.T @ right_sketch


def sketch_psi_cp(
    left_sketch: npt.NDArray,
    right_sketch: npt.NDArray,
    *,
    tensor: CPTensor,
    mu: int,
    **kwargs,
) -> npt.NDArray:
    cp_core = tensor.cores[mu]

    if left_sketch is None:
        Psi = np.einsum("ji,il->jl", cp_core, right_sketch)
        Psi = Psi.reshape((1,) + Psi.shape)
    elif right_sketch is None:
        Psi = np.einsum("il,kl->ik", left_sketch.T, cp_core)
        Psi = Psi.reshape(Psi.shape + (1,))
    else:
        Psi = np.einsum(
            "ij,kj,jm->ikm",
            left_sketch.T,
            cp_core,
            right_sketch,
            optimize="optimal",
        )
    return Psi



# %%
import numpy as np
from itertools import product
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.drm import TensorTrainDRM, DenseGaussianDRM
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor
from tt_sketch.sketch import SketchedTensorTrain, stream_sketch, hmt_sketch
from tt_sketch.utils import process_tt_rank
from experiment_base import (
    Experiment,
    experiment_stream_sketch,
    experiment_tt_round,
    experiment_hmt_sketch,
)

csv_filename = "results/dimension_scaling_guassian.csv"
experiment = Experiment(csv_filename)

seed = 179
tt_rank = 10
round_rank = 8
dim = 5
n_dims_list = np.arange(4, 13, dtype=int)
runs = range(30)
# %%
def tt_exp_decay(shape, tt_rank, min_svdval=-20, seed=None):
    tensor = TensorTrain.random(
        shape, rank=tt_rank, orthog=True, trim=True, seed=seed
    )
    for i, C in enumerate(tensor.cores):
        C_shape = C.shape
        left_mat_shape = (C_shape[0] * C.shape[1], C_shape[2])
        right_mat_shape = (C_shape[0], C_shape[1] * C.shape[2])
        if min(left_mat_shape) > min(right_mat_shape):
            mat_shape = left_mat_shape
        else:
            mat_shape = right_mat_shape
        C = C.reshape(mat_shape)
        U, S, Vt = np.linalg.svd(C, full_matrices=False)
        S = np.logspace(0, min_svdval, len(S))
        S = S * np.sqrt(min(mat_shape))
        C = (U @ np.diag(S) @ Vt).reshape(C_shape)
        tensor.cores[i] = C
    tensor.orthogonalize()
    return tensor


def tt_error_func(tt1, tt2, relative=True):
    if isinstance(tt1, SketchedTensorTrain):
        tt1 = tt1.to_tt()
    tt2_norm = tt2.norm()
    if tt2_norm == 0:
        return np.inf
    error = tt1.add(-tt2).norm()
    if relative:
        return error / tt2_norm
    return error


def make_tensor(n_dims, tt_rank, dim, seed=None, min_svdval=-20):
    shape = (dim,) * n_dims
    return tt_exp_decay(shape, tt_rank, seed=seed, min_svdval=min_svdval)



# %%
for n_dims in tqdm(n_dims_list):
    tensor = make_tensor(
        n_dims, tt_rank, dim, seed=seed + n_dims, min_svdval=-5
    )
    round_rank_trim = process_tt_rank(round_rank, tensor.shape, trim=True)
    right_rank = tuple(r*2 for r in round_rank_trim)
    experiment.do_experiment(
        tensor,
        "TT-SVD",
        experiment_tt_round,
        rank=round_rank_trim,
        n_dims=n_dims,
        error_func=tt_error_func,
    )
    for run in runs:
        experiment.do_experiment(
            tensor,
            "HMT",
            experiment_hmt_sketch,
            drm_type=DenseGaussianDRM,
            rank=round_rank_trim,
            run=run,
            n_dims=n_dims,
            error_func=tt_error_func,
        )
        experiment.do_experiment(
            tensor,
            "STTA",
            experiment_stream_sketch,
            drm_type=DenseGaussianDRM,
            left_rank=round_rank_trim,
            right_rank=right_rank,
            run=run,
            n_dims=n_dims,
            error_func=tt_error_func,
        )


# %%
import pandas as pd

df = pd.read_csv(csv_filename, index_col=0)
df.groupby(["name", "n_dims"]).median().reset_index()
n_dims_plot = df["n_dims"].unique()

df[df["name"] == "TT-SVD"]
# %%
from matplotlib.ticker import ScalarFormatter, NullFormatter, AutoMinorLocator

plt.figure(figsize=(8, 4))
marker_size = 4
markers = ["D", "s"]
jitter_dict = {
    "HMT": -0.07,
    "STTA": 0.07,
}
for name, marker in zip(("HMT", "STTA"), markers):
    df_subset = df[df["name"] == name]
    tt_svd = df[ (df["name"] == "TT-SVD")][
        "error"
    ].array
    # tt_svd = 1
    label_name = "TT-HMT" if name == "HMT" else name
    df_gb_median = df_subset.groupby("n_dims").median()
    error80 = df[df["name"] == name].groupby("n_dims")["error"].quantile(0.8)
    error20 = df[df["name"] == name].groupby("n_dims")["error"].quantile(0.2)
    plt.errorbar(
        df_gb_median.index + jitter_dict[name],
        df_gb_median["error"] / tt_svd,
        yerr=(
            (df_gb_median["error"] - error20) / tt_svd,
            (error80 - df_gb_median["error"]) / tt_svd,
        ),
        capsize=5,
        marker=marker,
        markersize=marker_size,
        label=label_name,
        linestyle="",
    )


ax = plt.gca()
ax.set_xticks(n_dims_plot)

plt.axhline(1, color="k", linestyle="--",alpha=0.7)

plt.legend()
plt.title("Scaling of error with number of modes")
plt.xlabel("Number of modes")
plt.ylabel("Normalized error")
plt.savefig(
    "results/plot-dimension-scaling-gaussian.pdf",
    transparent=True,
    bbox_inches="tight",
)

# %%



from tt_sketch.drm.dense_gaussian_drm import DenseGaussianDRM
from tt_sketch.drm.sparse_gaussian_drm import SparseGaussianDRM
from tt_sketch.drm.tensor_train_drm import TensorTrainDRM
from tt_sketch.drm.sparse_sign_drm import SparseSignDRM


ALL_DRM = (DenseGaussianDRM, SparseGaussianDRM, TensorTrainDRM, SparseSignDRM)



# %%
import numpy as np
from itertools import product
from tqdm import tqdm
import matplotlib.pyplot as plt
from tt_sketch.drm.tensor_train_drm import TensorTrainDRM
from tt_sketch.tensor import TensorTrain, TensorSum, Tensor
from tt_sketch.sketch import SketchedTensorTrain, stream_sketch, hmt_sketch
from experiment_base import (
    Experiment,
    experiment_stream_sketch,
    experiment_tt_round,
    experiment_hmt_sketch,
)

csv_filename = "results/dimension_scaling.csv"
experiment = Experiment(csv_filename)

seed = 179
tt_rank = 30
round_rank = 10
dim = 30
n_dims_list = 2 ** np.arange(2, 14)
runs = range(30)
# %%


def tt_exp_decay(shape, tt_rank, min_svdval=-20, seed=None):
    tensor = TensorTrain.random(
        shape, rank=tt_rank, orthog=True, trim=True, seed=seed
    )
    for i, C in enumerate(tensor.cores):
        C_shape = C.shape
        left_mat_shape = (C_shape[0] * C.shape[1], C_shape[2])
        right_mat_shape = (C_shape[0], C_shape[1] * C.shape[2])
        if min(left_mat_shape) > min(right_mat_shape):
            mat_shape = left_mat_shape
        else:
            mat_shape = right_mat_shape
        C = C.reshape(mat_shape)
        U, S, Vt = np.linalg.svd(C, full_matrices=False)
        S = np.logspace(0, min_svdval, len(S))
        S = S * np.sqrt(min(mat_shape))
        C = (U @ np.diag(S) @ Vt).reshape(C_shape)
        tensor.cores[i] = C
    return tensor


def tt_error_func(tt1, tt2, relative=True):
    if isinstance(tt1, SketchedTensorTrain):
        tt1 = tt1.to_tt()
    tt2_norm = tt2.norm()
    if tt2_norm == 0:
        return np.inf
    error = tt1.add(-tt2).norm()
    if relative:
        return error / tt2_norm
    return error


def make_tensor(n_dims, tt_rank, dim, seed=None):
    shape = (dim,) * n_dims
    return tt_exp_decay(shape, tt_rank, seed=seed)


# %%
for n_dims in tqdm(n_dims_list):
    tensor = make_tensor(n_dims, tt_rank, dim, seed=seed + n_dims)
    experiment.do_experiment(
        tensor,
        "TT-SVD",
        experiment_tt_round,
        rank=round_rank,
        n_dims=n_dims,
        error_func=tt_error_func,
    )
    experiment.do_experiment(
        tensor,
        "TT-SVD",
        experiment_tt_round,
        rank=round_rank - 1,
        n_dims=n_dims,
        error_func=tt_error_func,
    )
    experiment.do_experiment(
        tensor,
        "TT-SVD",
        experiment_tt_round,
        rank=round_rank - 2,
        n_dims=n_dims,
        error_func=tt_error_func,
    )
    for run in runs:
        experiment.do_experiment(
            tensor,
            "HMT",
            experiment_hmt_sketch,
            rank=round_rank,
            run=run,
            n_dims=n_dims,
            error_func=tt_error_func,
        )
        experiment.do_experiment(
            tensor,
            "STTA",
            experiment_stream_sketch,
            left_rank=round_rank,
            right_rank=round_rank * 2,
            run=run,
            n_dims=n_dims,
            error_func=tt_error_func,
        )


# %%
import pandas as pd

df = pd.read_csv(csv_filename, index_col=0)
df.groupby(["name", "n_dims"]).median().reset_index()
n_dims_plot = df["n_dims"].unique()


# %%
df["error"].array
# %%
from matplotlib.ticker import ScalarFormatter, NullFormatter, AutoMinorLocator

plt.figure(figsize=(8, 4))
marker_size = 4
markers = ["o", "o", "D", "s"]
jitter_dict = {
    "HMT": 1 / 1.05,
    "STTA": 1.05,
}
for name, marker in zip(
    ("TT-SVD, rank 8", "TT-SVD, rank 9", "HMT", "STTA"), markers
):
    if name == "TT-SVD, rank 8":
        df_subset = df[(df["rank"] == 8) & (df["name"] == "TT-SVD")]
    elif name == "TT-SVD, rank 9":
        df_subset = df[(df["rank"] == 9) & (df["name"] == "TT-SVD")]
    else:
        df_subset = df[df["name"] == name]
    tt_svd = df[(df["rank"] == 10) & (df["name"] == "TT-SVD")]["error"].array
    label_name = "TT-HMT" if name == "HMT" else name
    df_gb_median = df_subset.groupby("n_dims").median()
    if name.startswith("TT-SVD"):
        continue
        plt.plot(
            df_gb_median.index,
            df_gb_median["error"] / tt_svd,
            marker=marker,
            markersize=marker_size,
            label=label_name,
        )
    else:
        error80 = (
            df[df["name"] == name].groupby("n_dims")["error"].quantile(0.8)
        )
        error20 = (
            df[df["name"] == name].groupby("n_dims")["error"].quantile(0.2)
        )
        plt.errorbar(
            df_gb_median.index * jitter_dict[name],
            df_gb_median["error"] / tt_svd,
            yerr=(
                (df_gb_median["error"] - error20) / tt_svd,
                (error80 - df_gb_median["error"]) / tt_svd,
            ),
            capsize=5,
            marker=marker,
            markersize=marker_size,
            label=label_name,
            linewidth=1,
            # linestyle="",

        )

plt.xscale("log")

ax = plt.gca()
ax.set_xticks(n_dims_plot)
ax.xaxis.set_major_formatter(ScalarFormatter())
ax.xaxis.set_minor_formatter(NullFormatter())
ax.xaxis.set_tick_params(which="minor", size=0)
ax.xaxis.set_tick_params(which="minor", width=0)

ax.set_yticks(np.arange(2, 18 + 0.1, 2))
ax.yaxis.set_minor_locator(AutoMinorLocator(2))

plt.legend()
plt.title("Dependence of the error on the tensor order")
plt.xlabel("Order of tensor")
plt.ylabel("Normalized error")
plt.savefig(
    "results/plot-dimension-scaling.pdf", transparent=True, bbox_inches="tight"
)

# %%



from typing import Optional, Tuple, Union

import numpy as np
from tt_sketch.drm.fast_lazy_gaussian import inds_to_sparse_sign  # type: ignore
from tt_sketch.drm_base import CanSlice, handle_transpose
from tt_sketch.sketching_methods.abstract_methods import CansketchSparse
from tt_sketch.tensor import SparseTensor
from tt_sketch.utils import ArrayGenerator


class SparseSignDRM(CansketchSparse, CanSlice):
    """
    Sparse DRM where each row is a vector with fixed number of +/-1 entries.

    The number of nonzero entries are determined by ``num_non_zero_per_row``.
    Like ``SparseGaussianDRM``, entries are computed lazily/on-demand using a
    hashing algorithm.
    """

    def __init__(
        self,
        rank: Union[Tuple[int, ...], int],
        shape: Tuple[int, ...],
        transpose: bool,
        seed: Optional[int] = None,
        num_non_zero_per_row: Optional[Tuple[int, ...]] = None,
        **kwargs,
    ) -> None:
        super().__init__(rank, shape, transpose, seed=seed, **kwargs)
        if num_non_zero_per_row is None:
            num_non_zero_per_row = self.true_rank
        self.nnz = num_non_zero_per_row

    @handle_transpose
    def sketch_sparse(self, tensor: SparseTensor) -> ArrayGenerator:
        d = len(tensor.shape)
        for mu in range(d - 1):
            shape = tensor.shape[: mu + 1]
            sketch_seed = np.mod(
                mu + self.seed, 2**63, dtype=np.uint64
            )  # ensure safe casting to uint
            sketch_mat = inds_to_sparse_sign(
                tensor.indices[: mu + 1],
                shape,
                self.true_rank[mu],
                self.rank_min[mu],
                self.rank_max[mu],
                self.nnz[mu],
                sketch_seed,
            )
            yield sketch_mat.T



# %%
from operator import sub
from data_download_util import download_data
import os.path
from tqdm import tqdm
import numpy as np
import gzip
from tt_sketch.tensor import SparseTensor
from tt_sketch.sketched_tensor_train import SketchedTensorTrain
from tt_sketch.drm import (
    TensorTrainDRM,
    SparseSignDRM,
    SparseGaussianDRM,
)
from tt_sketch.sketching_methods.sparse_sketch import (
    sparse_sketch,
)

# File url for the NIPS dataset
nips_tensor = {
    "file_url": (
        "https://s3.us-east-2.amazonaws.com/frostt/frostt_data/nips/nips.tns.gz"
    ),
    "nnz": 3101609,
    "shape": (2482, 2862, 14036, 17),
}
uber_tensor = {
    "file_url": (
        "https://s3.us-east-2.amazonaws.com/frostt/frostt_data/uber-pickups/uber.tns.gz"
    ),
    "nnz": 3309490,
    "shape": (183, 24, 1140, 1717),
}
lbnl_tensor = {
    "file_url": (
        "https://s3.us-east-2.amazonaws.com/frostt/frostt_data/lbnl-network/lbnl-network.tns.gz"
    ),
    "nnz": 1698825,
    "shape": (1605, 4198, 1631, 4209, 868131),
}

matmul_tensor = {
    "file_url": (
        "https://s3.us-east-2.amazonaws.com/frostt/frostt_data/matrix-multiplication/matmul_5-5-5.tns.gz"
    ),
    "nnz": 5 * 5 * 5,
    "shape": (5 * 5, 5 * 5, 5 * 5),
}


def process_frostt_tensor(filepath, nnz, shape):
    with gzip.open(filepath) as f:
        lines = f.readlines()
        num_lines = nnz
        dim = len(lines[0].split(b" ")) - 1
        indices = np.zeros((num_lines, dim), dtype=np.int64)
        entires = np.zeros(num_lines, dtype=np.float64)
        for i, line in tqdm(
            enumerate(lines), total=nnz, desc="Processing file"
        ):
            split = line.split(b" ")
            indices[i] = [int(s) - 1 for s in split[:-1]]  # indices are 1-based
            entires[i] = float(split[-1])
    tensor = SparseTensor(shape, indices.T, entires)
    return tensor


def get_frostt_tensor(file_url, nnz, shape):
    filename = file_url.split("/")[-1]
    filepath = os.path.join("data", filename)
    npzpath = filepath.split(".gz")[0] + ".npz"
    try:
        npz_load = np.load(npzpath)
        tensor = SparseTensor(
            tuple(npz_load["shape"]), npz_load["indices"], npz_load["entries"]
        )
        return tensor
    except FileNotFoundError:
        download_data(file_url)
        tensor = process_frostt_tensor(filepath, nnz, shape)
        indices = tensor.indices
        entries = tensor.entries
        print("Saving processed file")
        np.savez_compressed(
            npzpath, indices=indices, entries=entries, shape=shape
        )
        print("Done!")
        return tensor


# %%
tensor_info = matmul_tensor
tensor = get_frostt_tensor(
    tensor_info["file_url"], tensor_info["nnz"], tensor_info["shape"]
)
ranks = range(2, 50)
errors = []


def compute_error(rank):
    Psi_cores, Omega_mats = sparse_sketch(
        tensor,
        rank,
        rank + 1,
        SparseGaussianDRM,
        SparseGaussianDRM,
        seed=179,
        pbar=None
    )
    stt = SketchedTensorTrain(None, None, Psi_cores, Omega_mats)
    tt = stt.to_tt()
    sample = np.random.choice(tensor.nnz, size=10000)
    inds = tensor.indices[:, sample]
    entr = tensor.entries[sample]

    rel_error = np.linalg.norm(tt.gather(inds) - entr) / np.linalg.norm(entr)
    # errors.append(rel_error)
    return rel_error


for rank in tqdm(ranks):
    errors.append(compute_error(rank))

# %%
import matplotlib.pyplot as plt

plt.plot(ranks, errors)
plt.yscale("log")

# %%



from __future__ import annotations

from abc import ABC
from copy import deepcopy
from typing import Any, Callable, Dict, Optional, Tuple, Union

from numpy import mod
from numpy.random import uniform

from tt_sketch.utils import TTRank, process_tt_rank


class DRM(ABC):
    """Implements an abstract dimension reduction matrix"""

    rank: Tuple[int, ...]  # The rank of the DRM. (Size of slice for blocked)
    rank_min: Tuple[int, ...]  # For blocked sketch, the start of rank slice
    rank_max: Tuple[int, ...]  # For blocked sketch, the end of rank slice
    true_rank: Tuple[int, ...]  # For blocked sketch, the true size before slice
    shape: Tuple[int, ...]  # shape of tensor
    transpose: bool  # If false, left sketch, if true right sketch
    seed: int

    def __init__(
        self,
        rank: TTRank,
        shape: Tuple[int, ...],
        transpose: bool,
        seed: Optional[int] = None,
        rank_min: Optional[Tuple[int, ...]] = None,
        rank_max: Optional[Tuple[int, ...]] = None,
        true_rank: Optional[Tuple[int, ...]] = None,
        **kwargs,
    ) -> None:
        self.transpose = transpose
        rank = process_tt_rank(rank, shape, trim=False)
        if true_rank is None:
            self.true_rank = rank
        else:
            self.true_rank = true_rank

        if rank_min is None:
            self.rank_min = (0,) * (len(shape) - 1)
        else:
            self.rank_min = rank_min
        if rank_max is None:
            self.rank_max = rank
        else:
            self.rank_max = rank_max

        if transpose:
            self.true_rank = self.true_rank[::-1]
            self.rank_min = self.rank_min[::-1]
            self.rank_max = self.rank_max[::-1]
        self.rank = tuple(
            r2 - r1 for r1, r2 in zip(self.rank_min, self.rank_max)
        )

        self.shape = shape
        if seed is None:
            seed = hash(uniform())
        seed = mod(seed, 2**32 - 1)
        self.seed = seed  # type: ignore

    @property
    def T(self):
        transposed = deepcopy(self)
        transposed.transpose = not self.transpose
        transposed.true_rank = transposed.true_rank[::-1]
        transposed.rank_min = transposed.rank_min[::-1]
        transposed.rank_max = transposed.rank_max[::-1]
        transposed.rank = transposed.rank[::-1]
        return transposed

    def __repr__(self) -> str:
        if self.transpose:
            direction = "Right"
        else:
            direction = "Left"
        return (
            f"<{direction} {self.__class__.__name__} of rank {self.rank}"
            f" and shape {self.shape} at {hex(id(self))}>"
        )


class CanSlice(DRM):
    """Mixin telling that the DRM supports slicing.

    Default implementation is inefficient, and in many cases should be
    overloaded."""

    def slice(
        self, start_rank: Tuple[int, ...], end_rank: Tuple[int, ...]
    ) -> DRM:
        if self.transpose:
            # If transpose, the rank will be reversed during init.
            # To keep rank the same, need to reverse it here as well.
            new_true_rank = self.true_rank[::-1]
        else:
            new_true_rank = self.true_rank
        return self.__class__(
            rank=self.rank,
            shape=self.shape,
            transpose=self.transpose,
            seed=self.seed,
            rank_min=start_rank,
            rank_max=end_rank,
            true_rank=new_true_rank,
        )


class CanIncreaseRank(CanSlice):
    """Mixin telling that the DRM supports rank increase and slicing.

    Standard implementations of increase_rank and slice are not optimal, and
    should be overloaded for good performance."""

    def increase_rank(self, new_rank: Tuple[int, ...]) -> DRM:
        return self.__class__(new_rank, self.shape, self.transpose, self.seed)


def handle_transpose(sketch: Callable) -> Callable:
    """Decorator to handle transpose of sketch.

    This way we only have to implement sketches on the left, and right-sketches
    are handled automatically."""

    def wrapper(self, tensor):
        if self.shape != tensor.shape:
            # Catch this shape mismatch early, because it can cause unexpected
            # behavior otherwise
            raise ValueError(
                f"""Shape {self.shape} of DRM doesn't match tensor's shape
                {tensor.shape}"""
            )
        if self.transpose:
            tensor = tensor.T

        sketching_mats = sketch(self, tensor)
        if self.transpose:
            sketching_mats = list(sketching_mats)[::-1]
        for mat in sketching_mats:
            yield mat

    return wrapper



import enum
import pytest
from sklearn.metrics import auc
from tt_sketch.tensor import (
    CPTensor,
    DenseTensor,
    SparseTensor,
    TensorTrain,
    TuckerTensor,
)
import numpy as np


def test_tt_round():
    shape = (5, 5, 5, 5)
    ndim = len(shape)
    tt = TensorTrain.random(shape, 2)
    new_cores = list(tt.cores)
    new_rank = 5
    for i in range(ndim - 1):
        r = tt.rank[i]
        Q = np.random.normal(size=(new_rank, r))
        Q, _ = np.linalg.qr(Q)
        new_cores[i] = np.einsum("ijk,kl->ijl", new_cores[i], Q.T)
        new_cores[i + 1] = np.einsum("ij,jkl->ikl", Q, new_cores[i + 1])
    tt2 = TensorTrain(new_cores)

    assert tt.error(tt2) < 1e-8

    tt3 = tt2.round(eps=1e-12)
    assert tt3.rank == tt.rank
    assert tt2.rank != tt.rank  # check for side effects
    assert tt.error(tt3) < 1e-8

    tt4 = tt2.round(max_rank=2)
    assert tt4.rank == tt.rank
    assert tt.error(tt4) < 1e-8


def test_tt_add():
    shape = (5, 5, 5, 5)
    rank = 2
    tt1 = TensorTrain.random(shape, rank)
    tt2 = TensorTrain.random(shape, rank)
    tt3 = tt1.add(tt2)
    assert tt3.rank == (4, 4, 4)
    assert (
        np.linalg.norm(tt3.to_numpy() - tt1.to_numpy() - tt2.to_numpy()) < 1e-8
    )
    tt0 = TensorTrain.zero(shape, rank)

    tt12 = tt1 + tt2
    tt120 = tt12 + tt0
    tt123 = tt12 + tt3
    tt312 = tt3 + tt12
    tt1212 = tt12 + tt12
    assert (
        np.linalg.norm(tt12.to_numpy() - tt1.to_numpy() - tt2.to_numpy()) < 1e-8
    )
    assert np.linalg.norm(tt12.to_numpy() - tt120.to_numpy()) < 1e-8
    assert (
        np.linalg.norm(
            tt1212.to_numpy() - 2 * tt1.to_numpy() - 2 * tt2.to_numpy()
        )
        < 1e-8
    )

    assert (
        np.linalg.norm(
            tt123.to_numpy() - 2 * tt1.to_numpy() - 2 * tt2.to_numpy()
        )
        < 1e-8
    )
    assert (
        np.linalg.norm(
            tt312.to_numpy() - 2 * tt1.to_numpy() - 2 * tt2.to_numpy()
        )
        < 1e-8
    )


def test_tt_dot():
    shape = (5, 5, 5, 5)
    rank = 2

    tt1 = TensorTrain.random(shape, rank)
    tt2 = TensorTrain.random(shape, rank)

    prod1 = tt1.dot(tt2)
    prod2 = tt1.to_numpy().reshape(-1).dot(tt2.to_numpy().reshape(-1))
    assert abs(prod1 - prod2) < 1e-8

    tt3 = tt1.orthogonalize()
    assert abs(np.linalg.norm(tt3[-1]) - tt3.norm()) < 1e-8


def make_cp(shape):
    return CPTensor.random(shape, 10)


def make_tt(shape):
    return TensorTrain.random(shape, 5)


def make_tucker(shape):
    return TuckerTensor.random(shape, 5)


def make_sparse(shape):
    return SparseTensor.random(shape, 5)


def make_dense(shape):
    return DenseTensor.random(shape)


class TensorType(enum.Enum):
    CP = enum.auto()
    TT = enum.auto()
    TUCKER = enum.auto()
    SPARSE = enum.auto()
    DENSE = enum.auto()


make_tensor_dispatch = {
    TensorType.CP: make_cp,
    TensorType.TT: make_tt,
    TensorType.TUCKER: make_tucker,
    TensorType.SPARSE: make_sparse,
    TensorType.DENSE: make_dense,
}


@pytest.mark.parametrize("tensor_type1", make_tensor_dispatch.keys())
@pytest.mark.parametrize("tensor_type2", make_tensor_dispatch.keys())
def test_dot(tensor_type1, tensor_type2):
    shape = (5, 6, 7, 4)
    tensor1 = make_tensor_dispatch[tensor_type1](shape)
    tensor2 = make_tensor_dispatch[tensor_type2](shape)
    dot_fast = tensor1 @ tensor2
    dot_slow = (
        tensor1.to_numpy().reshape(-1).dot(tensor2.to_numpy().reshape(-1))
    )
    assert abs(dot_fast - dot_slow) < 1e-8

    assert tensor1.to_numpy().shape == shape
    assert tensor2.to_numpy().shape == shape

    norm_fast = tensor1.norm()
    assert abs(norm_fast - np.linalg.norm(tensor1.to_numpy())) < 1e-8

    error_fast = tensor1.error(tensor2, fast=True)
    error_slow = np.linalg.norm(tensor1.to_numpy() - tensor2.to_numpy())
    assert abs(error_fast - error_slow) < 1e-8

    error_relative = tensor1.error(tensor2, relative=True)
    error_slow_relative = error_slow / tensor2.norm()
    assert abs(error_relative - error_slow_relative) < 1e-8

    error_inf = tensor1.error(0 * tensor2, relative=True)
    assert error_inf == np.inf

    error_rmse = tensor1.error(tensor2, rmse=True)
    error_rmse_slow = np.sqrt(
        np.mean((tensor1.to_numpy() - tensor2.to_numpy()) ** 2)
    )
    assert abs(error_rmse - error_rmse_slow) < 1e-8


def test_error_tt():
    shape = (5, 5, 5, 5)
    rank = 2
    tt1 = TensorTrain.random(shape, rank)
    tt2 = TensorTrain.random(shape, rank)
    error = tt1.error(tt2)
    assert abs(error - np.linalg.norm(tt1.to_numpy() - tt2.to_numpy())) < 1e-8

    tt3 = tt1 + 1e-10 * tt2
    error = tt1.error(tt3)
    assert abs(error - np.linalg.norm(tt1.to_numpy() - tt3.to_numpy())) < 1e-8


@pytest.mark.parametrize("tensor_type1", make_tensor_dispatch.keys())
@pytest.mark.parametrize("tensor_type2", make_tensor_dispatch.keys())
def test_arithmetic(tensor_type1, tensor_type2):
    shape = (5, 6, 7, 4)
    tensor1 = make_tensor_dispatch[tensor_type1](shape)
    tensor2 = make_tensor_dispatch[tensor_type2](shape)

    tensor3 = 2.3 * tensor1 - (1.5 * tensor2)
    assert (
        abs(
            np.linalg.norm(
                tensor3.to_numpy()
                - (2.3 * tensor1.to_numpy() - 1.5 * tensor2.to_numpy())
            )
        )
        < 1e-8
    )

    tensor4 = -tensor1 / 2 + (-1.5) * tensor2
    assert (
        abs(
            np.linalg.norm(
                tensor4.to_numpy()
                - (-tensor1.to_numpy() / 2 + (-1.5) * tensor2.to_numpy())
            )
        )
        < 1e-8
    )

    tensor5 = tensor3 + tensor4
    assert (
        abs(
            np.linalg.norm(
                tensor5.to_numpy() - (tensor3.to_numpy() + tensor4.to_numpy())
            )
        )
        < 1e-8
    )

    tensor6 = 1.4 * tensor3.T - 2 * tensor1.T
    assert (
        abs(
            np.linalg.norm(
                tensor6.to_numpy()
                - (1.4 * tensor3.to_numpy().T - 2 * tensor1.to_numpy().T)
            )
        )
        < 1e-8
    )

    tensor6 += tensor5.T
    assert (
        abs(
            np.linalg.norm(
                tensor6.to_numpy()
                - (1.4 * tensor3.to_numpy().T - 2 * tensor1.to_numpy().T+tensor5.to_numpy().T)
            )
        )
        < 1e-8
    )


# from tt_sketch.tt_gmres import (
#     tt_weighted_sum_sketched,
#     tt_weighted_sum_exact,
#     tt_gmres,
# )
import pytest
from tt_sketch.tt_gmres import MPO
from tt_sketch.tensor import TensorTrain
import numpy as np


def test_mpo_contract():
    shape = (4, 5, 6)

    mpo = MPO.random(3, shape, shape)

    mpo_dense = mpo.to_numpy()

    tt = TensorTrain.random(shape, 2)

    mpo_dense_t = mpo_dense.transpose((1, 0, 3, 2, 5, 4))
    assert np.linalg.norm(mpo_dense - mpo_dense_t) < 1e-12  # assert symmetry

    tt_dense = tt.to_numpy()

    tt_mpo1 = np.einsum("ijk,iajbkc", tt_dense, mpo_dense)
    tt_mpo2 = mpo(tt).to_numpy()

    assert np.linalg.norm(tt_mpo1 - tt_mpo2) < 1e-12


# @pytest.mark.parametrize("method", ["sketched", "exact"])
# def test_tt_weighted_sum(method):
#     shape = (4, 5, 6)
#     n_elems = 10
#     x0 = TensorTrain.random(shape, 3)
#     tt_list = [TensorTrain.random(shape, 3) for _ in range(n_elems)]
#     coeffs = np.random.normal(size=n_elems)

#     dense = x0.to_numpy()
#     for tt, coeff in zip(tt_list, coeffs):
#         dense += coeff * tt.to_numpy()

#     if method == "exact":
#         tt_sum = tt_weighted_sum_exact(x0, coeffs, tt_list, 1e-6, 6)
#     if method == "sketched":
#         tt_sum = tt_weighted_sum_sketched(x0, coeffs, tt_list, 1e-6, 6)
#     assert np.linalg.norm(tt_sum.to_numpy() - dense) < 1e-8


# def test_tt_gmres():
#     shape = (3, 3, 3)
#     mpo = MPO.random(4, shape, shape)
#     tt_true = TensorTrain.random(shape, 2)
#     tt = mpo(tt_true)

#     rank = 6
#     tt_approx, _ = tt_gmres(mpo, tt, rank, maxiter=100, tolerance=1e-8)
#     assert mpo(tt_approx).mse_error(tt) / tt.norm() < 1e-6



from __future__ import annotations

from typing import Optional, Tuple, Union

import numpy as np
from numpy.random import SeedSequence
from tt_sketch.drm_base import CanIncreaseRank, handle_transpose
from tt_sketch.sketching_methods.abstract_methods import (
    CansketchDense,
    CansketchSparse,
    CansketchTT,
)
from tt_sketch.tensor import DenseTensor, SparseTensor, TensorTrain
from tt_sketch.utils import ArrayGenerator, ArrayList, random_normal


class DenseGaussianDRM(
    CansketchTT, CansketchSparse, CansketchDense, CanIncreaseRank
):
    """Dense Gaussian DRM.

    The DRM is stored as a list of matrices, one for each mode."""

    sketching_mats: ArrayList

    def __init__(
        self,
        rank: Union[Tuple[int, ...], int],
        shape: Tuple[int, ...],
        transpose: bool,
        seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(rank, shape, transpose, seed=seed, **kwargs)

        self.sketching_mats = []
        dim_prod = 1
        shape_sketch = self.shape
        if transpose:
            shape_sketch = shape_sketch[::-1]

        # seq = SeedSequence(self.seed)
        # seeds = seq.generate_state(len(self.true_rank))
        for mu, (r, n) in enumerate(
            zip(self.true_rank, shape_sketch[:-1])
        ):
            dim_prod *= n

            # TODO: make this correctly work with rank_increase and random_normal
            np.random.seed(seed)
            seed_offset = hash(np.random.uniform(0, dim_prod))
            np.random.seed(np.mod(self.seed + seed_offset, 2**32 - 1))

            # sketching_mat = random_normal(shape=(r, dim_prod), seed=seed)
            sketching_mat = np.random.normal(size=(r, dim_prod))
            sketching_mat = sketching_mat[self.rank_min[mu] : self.rank_max[mu]]
            self.sketching_mats.append(sketching_mat)

    @handle_transpose
    def sketch_sparse(self, tensor: SparseTensor) -> ArrayGenerator:
        d = len(tensor.shape)
        for mu in range(d - 1):
            shape = tensor.shape[: mu + 1]
            inds = tensor.indices[: mu + 1]
            inds = np.ravel_multi_index(inds, shape)  # type: ignore
            sketching_vec = self.sketching_mats[mu][:, inds]
            yield sketching_vec

    @handle_transpose
    def sketch_tt(self, tensor: TensorTrain) -> ArrayGenerator:
        r"""Contract sketching matrix with :math:`Phi_{\leq \mu}`"""
        partial_contracts = tensor.partial_dense("lr")
        for sm, pc in zip(self.sketching_mats, partial_contracts):
            sketch = (sm @ pc).T
            yield sketch

    @handle_transpose
    def sketch_dense(self, tensor: DenseTensor) -> ArrayGenerator:
        for mat in self.sketching_mats:
            yield mat



"""Interface for the streaming and orthogonal sketching algorithms"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple, Type

import numpy as np
import numpy.typing as npt

from tt_sketch.drm import (
    ALL_DRM,
    DenseGaussianDRM,
    SparseGaussianDRM,
    TensorTrainDRM,
)
from tt_sketch.drm_base import DRM, CanIncreaseRank, CanSlice
from tt_sketch.sketch_container import SketchContainer
from tt_sketch.sketch_dispatch import SketchMethod, general_sketch
from tt_sketch.sketching_methods.abstract_methods import (
    CansketchCP,
    CansketchDense,
    CansketchSparse,
    CansketchTT,
)
from tt_sketch.tensor import Tensor, TensorTrain
from tt_sketch.utils import (
    ArrayList,
    TTRank,
    left_mul_pinv,
    process_tt_rank,
    right_mul_pinv,
)

DEFAULT_DRM = {
    CansketchDense: DenseGaussianDRM,
    CansketchSparse: SparseGaussianDRM,
    CansketchTT: TensorTrainDRM,
    CansketchCP: TensorTrainDRM,
}

BlockedSketch = Dict[Tuple[int, int], SketchContainer]


def hmt_sketch(
    tensor: Tensor,
    rank: TTRank,
    seed: Optional[int] = None,
    drm_type: Optional[Type[DRM]] = None,
    drm: Optional[DRM] = None,
    return_drm: bool = False,
) -> TensorTrain:
    """
    Perform an orthogonal sketch of a tensor
    """
    d = len(tensor.shape)

    if seed is None:
        seed = np.mod(hash(np.random.uniform()), 2**32)

    if drm is None:
        if drm_type is None:
            drm_type = TensorTrainDRM
        rank = process_tt_rank(rank, tensor.shape, trim=True)
        drm = drm_type(rank, transpose=True, shape=tensor.shape, seed=seed)
    else:
        if tuple(drm.rank[::-1]) != rank:
            raise ValueError(
                f"Right rank {rank} does not match the rank of the DRM "
                f"{drm.rank}."
            )

    sketch = general_sketch(tensor, None, drm, method=SketchMethod.hmt)

    sketched = TensorTrain(sketch.Psi_cores)
    if return_drm:  # this really is mostly for testing purposes
        return sketched, drm, right_drm  # type: ignore
    else:
        return sketched


def orthogonal_sketch(
    tensor: Tensor,
    left_rank: TTRank,
    right_rank: TTRank,
    seed: Optional[int] = None,
    left_drm_type: Optional[Type[DRM]] = None,
    right_drm_type: Optional[Type[DRM]] = None,
    left_drm: Optional[DRM] = None,
    right_drm: Optional[DRM] = None,
    return_drm: bool = False,
) -> TensorTrain:
    """
    Perform an orthogonal sketch of a tensor
    """
    d = len(tensor.shape)

    right_rank_bigger = bool(np.all(np.array(left_rank) < np.array(right_rank)))
    if not right_rank_bigger:
        raise ValueError(
            f"The right rank needs to be larger than the left rank. "
            f"Left rank: {left_rank}, "
            f"right rank: {right_rank}"
        )

    if seed is None:
        seed = np.mod(hash(np.random.uniform()), 2**32)

    if left_drm is None:
        if left_drm_type is None:
            if right_drm_type is not None:
                left_drm_type = right_drm_type
            else:
                left_drm_type = TensorTrainDRM
        left_rank = process_tt_rank(left_rank, tensor.shape, trim=True)
        left_drm = left_drm_type(
            left_rank, transpose=False, shape=tensor.shape, seed=seed
        )
    else:
        if left_drm.rank != left_rank:
            raise ValueError(
                f"Left rank {left_rank} does not match the rank of the DRM "
                f"{left_drm.rank}."
            )

    if right_drm is None:
        if right_drm_type is None:
            if left_drm_type is not None:
                right_drm_type = left_drm_type
            else:
                right_drm_type = TensorTrainDRM
        right_rank = process_tt_rank(right_rank, tensor.shape, trim=False)
        right_seed = np.mod(seed + hash(str(d)), 2**32)
        right_drm = right_drm_type(
            right_rank, transpose=True, shape=tensor.shape, seed=right_seed
        )
    else:
        if tuple(right_drm.rank[::-1]) != right_rank:
            raise ValueError(
                f"Right rank {right_rank} does not match the rank of the DRM "
                f"{right_drm.rank}."
            )

    sketch = general_sketch(
        tensor, left_drm, right_drm, method=SketchMethod.orthogonal
    )

    sketched = TensorTrain(sketch.Psi_cores)
    if return_drm:  # this really is mostly for testing purposes
        return sketched, left_drm, right_drm  # type: ignore
    else:
        return sketched


def stream_sketch(
    tensor: Tensor,
    left_rank: TTRank,
    right_rank: TTRank,
    seed: Optional[int] = None,
    left_drm_type: Optional[Type[DRM]] = None,
    right_drm_type: Optional[Type[DRM]] = None,
    left_drm: Optional[DRM] = None,
    right_drm: Optional[DRM] = None,
    return_drm: bool = False,
) -> SketchedTensorTrain:
    """
    Perform a streaming sketch of a tensor
    """
    d = len(tensor.shape)

    left_rank_bigger = bool(np.all(np.array(left_rank) > np.array(right_rank)))
    right_rank_bigger = bool(np.all(np.array(left_rank) < np.array(right_rank)))
    if not left_rank_bigger and not right_rank_bigger:
        raise ValueError(
            f"Left ranks or right ranks must be conistently larger or smaller "
            f"than the other. Left rank: {left_rank}, "
            f"right rank: {right_rank}"
        )

    if seed is None:
        seed = np.mod(hash(np.random.uniform()), 2**32)

    if left_drm is None:
        if left_drm_type is None:
            if right_drm_type is not None:
                left_drm_type = right_drm_type
            else:
                left_drm_type = TensorTrainDRM
        left_rank = process_tt_rank(
            left_rank, tensor.shape, trim=right_rank_bigger
        )
        left_drm = left_drm_type(
            left_rank, transpose=False, shape=tensor.shape, seed=seed
        )
    else:
        if left_drm.rank != left_rank:
            raise ValueError(
                f"Left rank {left_rank} does not match the rank of the DRM "
                f"{left_drm.rank}."
            )

    if right_drm is None:
        if right_drm_type is None:
            if left_drm_type is not None:
                right_drm_type = left_drm_type
            else:
                right_drm_type = TensorTrainDRM
        right_rank = process_tt_rank(
            right_rank, tensor.shape, trim=left_rank_bigger
        )
        right_seed = np.mod(seed + hash(str(d)), 2**32)
        right_drm = right_drm_type(
            right_rank, transpose=True, shape=tensor.shape, seed=right_seed
        )
    else:
        if tuple(right_drm.rank[::-1]) != right_rank:
            raise ValueError(
                f"Right rank {right_rank} does not match the rank of the DRM "
                f"{right_drm.rank}."
            )

    sketch = general_sketch(
        tensor, left_drm, right_drm, method=SketchMethod.streaming
    )

    sketched = SketchedTensorTrain(sketch, left_drm, right_drm)
    if return_drm:  # this really is mostly for testing purposes
        return sketched, left_drm, right_drm  # type: ignore
    else:
        return sketched


@dataclass
class SketchedTensorTrain(Tensor):
    """
    Container for storing the output of the streaming sketch

    Stores the result of the sketch as well as the DRMs used for the sketching.
    Can be cheaply converted to a tensor train, or the sketch can be efficiently
    updated using the ``__add__`` method.
    """

    sketch_: SketchContainer
    left_drm: DRM
    right_drm: DRM

    @property
    def left_rank(self) -> Tuple[int, ...]:
        return self.left_drm.rank

    @property
    def right_rank(self) -> Tuple[int, ...]:
        return self.right_drm.rank[::-1]

    @property
    def Psi_cores(self) -> ArrayList:
        return self.sketch_.Psi_cores

    @property
    def size(self) -> int:
        total_Psi_size = sum(Psi.size for Psi in self.Psi_cores)
        total_Omega_size = sum(Omega.size for Omega in self.Omega_mats)
        return total_Psi_size + total_Omega_size

    @property
    def Omega_mats(self) -> ArrayList:
        return self.sketch_.Omega_mats

    def __post_init__(self):
        self.shape = self.sketch_.shape

    def C_cores(self, direction="auto") -> ArrayList:
        return assemble_sketched_tt(self.sketch_, direction=direction)

    @property
    def T(self) -> SketchedTensorTrain:
        new_sketch = self.sketch_.T
        return self.__class__(new_sketch, self.right_drm.T, self.left_drm.T)

    def to_tt(self) -> TensorTrain:
        return TensorTrain(self.C_cores())

    def to_numpy(self) -> npt.NDArray[np.float64]:
        return self.to_tt().to_numpy()

    def __repr__(self) -> str:
        return (
            f"<Sketched tensor train of shape {self.shape} with left-rank "
            f"{self.left_rank} and right-rank {self.right_rank} "
            f"at {hex(id(self))}>"
        )

    def __add__(self, other: Tensor) -> SketchedTensorTrain:
        other_sketch = stream_sketch(
            other,
            self.left_rank,
            self.right_rank,
            left_drm=self.left_drm,
            right_drm=self.right_drm,
        )
        new_sketch = self.sketch_ + other_sketch.sketch_
        return self.__class__(new_sketch, self.left_drm, self.right_drm)

    def increase_rank(
        self,
        tensor: Tensor,
        new_left_rank: TTRank,
        new_right_rank: TTRank,
    ) -> SketchedTensorTrain:
        """Increase the rank of the approximation by performing a new sketch.

        Requires DRM with support for the ``CanIncreaseRank`` protocol, which
        currently is only supported by ``SparseGaussianDRM``.
        """
        new_left_rank = process_tt_rank(new_left_rank, tensor.shape, trim=False)
        new_right_rank = process_tt_rank(
            new_right_rank, tensor.shape, trim=False
        )
        for drm in (self.left_drm, self.right_drm):
            if not isinstance(drm, CanSlice):
                drm_name = drm.__class__.__name__
                raise ValueError(
                    f"Increasing rank is not supported for DRM {drm_name}"
                )

        n_dims = len(tensor.shape)
        left_rank_slices = [
            (0,) * (n_dims - 1),
            self.left_drm.rank,
            new_left_rank,
        ]
        right_rank_slices = [
            (0,) * (n_dims - 1),
            self.right_drm.rank[::-1],
            new_right_rank,
        ]
        left_drm = self.left_drm.increase_rank(new_left_rank)  # type: ignore
        right_drm = self.right_drm.increase_rank(new_right_rank)  # type: ignore

        sketch_dict = _blocked_stream_sketch_components(
            tensor,
            left_drm,
            right_drm,
            left_rank_slices,
            right_rank_slices,
            excluded_entries=[(0, 0)],
        )

        sketch_dict[(0, 0)] = self.sketch_
        sketch = _assemble_blocked_stream_sketches(
            left_rank_slices, right_rank_slices, tensor.shape, sketch_dict
        )

        return self.__class__(sketch, left_drm, right_drm)

    def __mul__(self, other: float) -> SketchedTensorTrain:
        return self.__class__(
            self.sketch_ * other, self.left_drm, self.right_drm
        )

    def dot(self, other: Tensor, reverse=False) -> float:
        return self.to_tt().dot(other, reverse)


def _blocked_stream_sketch_components(
    tensor: Tensor,
    left_rm: CanSlice,
    right_drm: CanSlice,
    left_rank_slices: List[Tuple[int, ...]],
    right_rank_slices: List[Tuple[int, ...]],
    excluded_entries: Optional[Sequence[Tuple[int, int]]] = None,
) -> BlockedSketch:
    if excluded_entries is None:
        excluded_entries = []
    block_left_sketches = [
        left_rm.slice(rank1, rank2)
        for rank1, rank2 in zip(left_rank_slices[:-1], left_rank_slices[1:])
    ]
    block_right_sketches = [
        right_drm.slice(rank1, rank2)
        for rank1, rank2 in zip(right_rank_slices[:-1], right_rank_slices[1:])
    ]

    # Compute all the sketches
    sketch_dict = {}
    for i, left_sketch_slice in enumerate(block_left_sketches):
        for j, right_sketch_slice in enumerate(block_right_sketches):
            if (i, j) in excluded_entries:
                continue
            sketch_block = general_sketch(
                tensor,
                left_sketch_slice,
                right_sketch_slice,
                method=SketchMethod.streaming,
            )
            sketch_dict[(i, j)] = sketch_block

    return sketch_dict


def assemble_sketched_tt(
    sketch: SketchContainer,
    direction="auto",
) -> ArrayList:
    """Reconstructs a TT from a sketch, using Psi and Omega matrices."""
    tt_cores = []
    if direction == "auto":
        left_rank_bigger = np.all(
            np.array(sketch.left_rank) > np.array(sketch.right_rank)
        )
        direction = "left" if left_rank_bigger else "right"

    if direction == "right":
        for Psi, Omega in zip(sketch.Psi_cores[:-1], sketch.Omega_mats):
            Psi_shape = Psi.shape
            Psi_mat = Psi.reshape(Psi_shape[0] * Psi_shape[1], Psi_shape[2])
            try:
                Psi_Omega_pinv = right_mul_pinv(Psi_mat, Omega)
            except ValueError:
                print(Psi.shape, Omega.shape)
                raise
            core = Psi_Omega_pinv.reshape(
                Psi_shape[0], Psi_shape[1], Omega.shape[0]
            )
            tt_cores.append(core)
        tt_cores.append(sketch.Psi_cores[-1])
    elif direction == "left":
        tt_cores.append(sketch.Psi_cores[0])
        for Psi, Omega in zip(sketch.Psi_cores[1:], sketch.Omega_mats):
            Psi_shape = Psi.shape
            Psi_mat = Psi.reshape(Psi_shape[0], Psi_shape[1] * Psi_shape[2])
            try:
                Omega_pinv_Psi = left_mul_pinv(Omega, Psi_mat)
            except ValueError:
                print(Psi.shape, Omega.shape)
                raise
            core = Omega_pinv_Psi.reshape(
                Omega.shape[1], Psi_shape[1], Psi_shape[2]
            )
            tt_cores.append(core)
    else:
        raise ValueError(f"Unknown direction {direction}")

    return tt_cores


def _assemble_blocked_stream_sketches(
    left_rank_slices: List[Tuple[int, ...]],
    right_rank_slices: List[Tuple[int, ...]],
    shape: Tuple[int, ...],
    sketch_dict: BlockedSketch,
) -> SketchContainer:
    left_rank = tuple(left_rank_slices[-1])
    right_rank = tuple(right_rank_slices[-1])

    sketch = SketchContainer.zero(shape, left_rank, right_rank)
    for (i, j), sketch_block in sketch_dict.items():
        left_rank1 = (0,) + left_rank_slices[i]
        left_rank2 = (1,) + left_rank_slices[i + 1]
        right_rank1 = right_rank_slices[j] + (0,)
        right_rank2 = right_rank_slices[j + 1] + (1,)
        for mu, Psi in enumerate(sketch_block.Psi_cores):
            sketch.Psi_cores[mu][
                left_rank1[mu] : left_rank2[mu],
                :,
                right_rank1[mu] : right_rank2[mu],
            ] = Psi
        for mu, Omega in enumerate(sketch_block.Omega_mats):
            sketch.Omega_mats[mu][
                left_rank1[mu + 1] : left_rank2[mu + 1],
                right_rank1[mu] : right_rank2[mu],
            ] = Omega

    return sketch


def get_drm_capabilities():
    """List what all the DRMs are capable of"""
    all_capabilities = {}
    for drm in ALL_DRM:
        drm_capabilities = {}
        for capability in (
            CanSlice,
            CanIncreaseRank,
            CansketchSparse,
            CansketchDense,
            CansketchTT,
        ):
            drm_capabilities[capability.__name__] = issubclass(drm, capability)
        all_capabilities[drm.__name__] = drm_capabilities
    return all_capabilities


def blocked_stream_sketch(
    tensor: Tensor,
    left_drm: CanSlice,
    right_drm: CanSlice,
    left_rank_slices: List[Tuple[int, ...]],
    right_rank_slices: List[Tuple[int, ...]],
) -> SketchContainer:
    """Do a blocked sketch.

    It's use is mainly theoretical, since this this would only be faster in a
    distributed setting (which isn't properly supported).
    """
    for drm in (left_drm, right_drm):
        if not isinstance(drm, CanSlice):
            drm_name = drm.__class__.__name__
            raise ValueError(f"Blocked sketch not supported for DRM {drm_name}")

    sketch_dict = _blocked_stream_sketch_components(
        tensor,
        left_drm,
        right_drm,
        left_rank_slices,
        right_rank_slices,
    )

    sketch = _assemble_blocked_stream_sketches(
        left_rank_slices,
        right_rank_slices,
        tensor.shape,
        sketch_dict,
    )

    return sketch



from typing import Optional, Tuple, Union

import numpy as np
from tt_sketch.drm_base import CanSlice, handle_transpose
from tt_sketch.sketching_methods.abstract_methods import (
    CansketchCP,
    CansketchDense,
    CansketchSparse,
    CansketchTT,
    CanSketchTucker,
)
from tt_sketch.tensor import (
    CPTensor,
    DenseTensor,
    SparseTensor,
    TensorTrain,
    TuckerTensor,
)
from tt_sketch.utils import ArrayGenerator, ArrayList


# TODO: Store DRM as a tensor.TensorTrain
class TensorTrainDRM(
    CansketchSparse,
    CansketchTT,
    CansketchCP,
    CanSlice,
    CansketchDense,
    CanSketchTucker,
):
    """
    Tensor train DRM. Sketches with partial contractions of a fixed TT.
    """

    cores: ArrayList

    def __init__(
        self,
        rank: Union[Tuple[int, ...], int],
        shape: Tuple[int, ...],
        transpose: bool,
        seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(rank, shape, transpose, seed=seed, **kwargs)
        if transpose:
            tt_shape = self.shape[::-1]
            tt_rank = self.true_rank
        else:
            tt_shape = self.shape
            tt_rank = self.true_rank
        if "cores" not in kwargs:
            tt = TensorTrain.random(
                tt_shape, tt_rank, self.seed, norm_goal="norm-preserve"
            )
            self.cores = tt.cores[:-1]
        else:
            self.cores = kwargs["cores"]

    @handle_transpose
    def sketch_sparse(self, tensor: SparseTensor) -> ArrayGenerator:
        for mu, core in enumerate(self.cores):
            core_slice = core[:, tensor.indices[mu], :]
            if mu == 0:
                lr_contract = core_slice.reshape(core_slice.shape[1:])
            else:
                lr_contract = np.einsum("ijk,ji->jk", core_slice, lr_contract)
            sketch_mat = lr_contract[:, self.rank_min[mu] : self.rank_max[mu]]
            yield sketch_mat.T

    @handle_transpose
    def sketch_tt(self, tensor: TensorTrain) -> ArrayGenerator:
        n_dims = len(self.shape)

        for mu in range(n_dims - 1):
            tensor_core = tensor.cores[mu]
            drm_core = self.cores[mu]
            if mu == 0:
                lr_contract = np.einsum("ijk,ijl->kl", tensor_core, drm_core)
            else:
                lr_contract = np.einsum(
                    "ij,ikl,jkm->lm",
                    lr_contract,
                    tensor_core,
                    drm_core,
                    optimize="optimal",
                )
            yield lr_contract[:, self.rank_min[mu] : self.rank_max[mu]]

    @handle_transpose
    def sketch_cp(self, tensor: CPTensor) -> ArrayGenerator:
        n_dims = len(self.shape)

        for mu in range(n_dims - 1):
            tensor_core = tensor.cores[mu]
            drm_core = self.cores[mu]
            if mu == 0:
                lr_contract = np.einsum("ij,lik->jk", tensor_core, drm_core)
            else:
                lr_contract = np.einsum(
                    "ij,ki,jkl->il",
                    lr_contract,
                    tensor_core,
                    drm_core,
                    optimize="optimal",
                )
            yield lr_contract[:, self.rank_min[mu] : self.rank_max[mu]]

    @handle_transpose
    def sketch_dense(self, tensor: DenseTensor) -> ArrayGenerator:
        n_dims = len(self.shape)
        partial_contraction = self.cores[0].reshape(-1, self.cores[0].shape[-1])
        yield partial_contraction.T
        for mu in range(1, n_dims - 1):
            core = self.cores[mu]
            partial_contraction = np.einsum(
                "ij,jkl->ikl", partial_contraction, core
            )
            partial_contraction = partial_contraction.reshape(
                -1, partial_contraction.shape[-1]
            )
            yield partial_contraction.T

    @handle_transpose
    def sketch_tucker(self, tensor: TuckerTensor) -> ArrayGenerator:
        n_dims = len(self.shape)
        partial_contraction = np.einsum(
            "ijk,jl->ilk", self.cores[0], tensor.factors[0].T
        )
        partial_contraction = partial_contraction.reshape(
            tensor.rank[0], self.rank[0]
        )
        yield partial_contraction

        for mu in range(1, n_dims - 1):
            core_reduced = np.einsum(
                "jkl,km->jml", self.cores[mu], tensor.factors[mu].T
            )
            partial_contraction = np.einsum(
                "ij,jml->iml", partial_contraction, core_reduced
            )
            partial_contraction = partial_contraction.reshape(
                -1, partial_contraction.shape[-1]
            )
            yield partial_contraction



"""Implements TT-SVD algorithm"""
from typing import Optional

import numpy as np

from tt_sketch.tensor import Tensor, TensorTrain
from tt_sketch.utils import TTRank, matricize, process_tt_rank


def tt_svd(tensor: Tensor, rank: Optional[TTRank] = None) -> TensorTrain:
    """Compute the TT-SVD of an array in a left-to-right sweep"""
    tensor_numpy = tensor.to_numpy()
    shape = tensor.shape
    n_dims = len(shape)
    if rank is None:
        rank = (np.prod(shape, dtype=int),) * (n_dims - 1)
    rank = process_tt_rank(rank, shape, trim=True)
    new_rank = list(rank)
    tt_cores = []

    tensor_mat = matricize(tensor_numpy, 0)
    U, S, V = np.linalg.svd(tensor_mat, full_matrices=False)
    r = max(min(U.shape[1], new_rank[0]), 1)
    new_rank[0] = r
    tt_core = U[:, :r]
    tt_core = tt_core.reshape(1, shape[0], r)
    tt_cores.append(tt_core)

    tensor_compressed = (np.diag(S[:r]) @ V[:r, :]).reshape(
        (r,) + tensor.shape[1:]
    )
    for mu in range(1, n_dims - 1):
        tensor_mat = matricize(tensor_compressed, (0, 1), mat_shape=True)
        U, S, V = np.linalg.svd(tensor_mat, full_matrices=False)
        r = max(min(U.shape[1], new_rank[mu]), 1)
        new_rank[mu] = r
        tt_core = U[:, :r].reshape(new_rank[mu - 1], shape[mu], r)
        tt_cores.append(tt_core)
        tensor_compressed = np.diag(S[:r]) @ V[:r, :]
        tensor_compressed = tensor_compressed.reshape(
            (r,) + tensor.shape[mu + 1 :]
        )

    tt_core = tensor_compressed.reshape(
        new_rank[n_dims - 2], shape[n_dims - 1], 1
    )
    tt_cores.append(tt_core)

    return TensorTrain(tt_cores)



from typing import Optional

import numpy as np
import numpy.typing as npt
from tt_sketch.tensor import TuckerTensor
from tt_sketch.utils import matricize


def sketch_omega_tucker(
    left_sketch: npt.NDArray,
    right_sketch: npt.NDArray,
    *,
    tensor: TuckerTensor,
    mu: int,
    **kwargs
):
    mode = tuple(range(mu + 1))
    core_mat = matricize(tensor.core, mode, mat_shape=True)
    return left_sketch.T @ core_mat @ right_sketch


def sketch_psi_tucker(
    left_sketch: Optional[npt.NDArray],
    right_sketch: Optional[npt.NDArray],
    *,
    tensor: TuckerTensor,
    mu: int,
    **kwargs
):
    left_dim = left_sketch.shape[0] if left_sketch is not None else 1
    right_dim = right_sketch.shape[0] if right_sketch is not None else 1
    core_ord3 = tensor.core.reshape(left_dim, tensor.rank[mu], right_dim)
    if left_sketch is None:
        Psi = np.einsum("ijk,kl->ijl", core_ord3, right_sketch)  # type: ignore
    elif right_sketch is None:
        Psi = np.einsum("ij,jkl->ikl", left_sketch.T, core_ord3)
    else:
        Psi = np.einsum(
            "ij,jkl,lm->ikm",
            left_sketch.T,
            core_ord3,
            right_sketch,
            optimize="optimal",
        )
    Psi = np.einsum("ijk,jl->ilk", Psi, tensor.factors[mu])
    return Psi



from typing import Optional, Tuple, Union

import numpy as np
from tt_sketch.drm.fast_lazy_gaussian import inds_to_normal  # type: ignore
from tt_sketch.drm_base import CanIncreaseRank, handle_transpose
from tt_sketch.sketching_methods.abstract_methods import CansketchSparse
from tt_sketch.tensor import SparseTensor
from tt_sketch.utils import ArrayGenerator


class SparseGaussianDRM(CansketchSparse, CanIncreaseRank):
    """'Sparse' Gaussian DRM

    Mathematically equivalent ``DenseGaussianDRM``, but entries of the DRM
    are computed lazily/on-demand using a hashing algorithm. This makes
    it computationally feasible for very sparse tensors.
    """

    def __init__(
        self,
        rank: Union[Tuple[int, ...], int],
        shape: Tuple[int, ...],
        transpose: bool,
        seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(rank, shape, transpose, seed=seed, **kwargs)

    @handle_transpose
    def sketch_sparse(self, tensor: SparseTensor) -> ArrayGenerator:
        d = len(tensor.shape)
        for mu in range(d - 1):
            shape = tensor.shape[: mu + 1]
            sketch_seed = np.mod(
                mu + self.seed, 2**63, dtype=np.uint64
            )  # ensure safe casting to uint
            sketch_mat = inds_to_normal(
                tensor.indices[: mu + 1],
                shape,
                self.rank_min[mu],
                self.rank_max[mu],
                sketch_seed,
            )
            yield sketch_mat.T



import sys

import scipy.stats
import numpy as np
import pytest

from tt_sketch.drm.fast_lazy_gaussian import (
    hash_int_c,
    inds_to_sparse_sign,
    inds_to_normal,
    _inds_to_rand_double,
)

# TODO: test cases where r_min > 0


def eight_bytes_to_bits(input_vals):
    """Convert 64-bit arrays to array of bits."""
    shape = input_vals.shape
    vals = np.copy(input_vals).reshape(-1)
    vals = np.frombuffer(vals, dtype=np.uint64)
    all_bytes = []
    for i in range(8):
        bytes_i = (vals << 56 - 8 * i) >> 56
        vals -= bytes_i << 8 * i
        all_bytes.append(bytes_i.astype(np.uint8))

    all_bytes = np.stack(all_bytes)
    bits = np.unpackbits(all_bytes, axis=0)
    bits = bits.reshape((64,) + shape)
    return bits


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("size", (1, 3))
def test_permutation_invariance(n_dims, size):
    """Tests if `inds_to_normal` works as expected"""
    N = 1000
    shape = (10,) * n_dims
    tot_dim = np.prod(shape)
    N = min(N, tot_dim)

    inds_flat = np.random.choice(tot_dim, size=N)
    inds = np.stack(np.unravel_index(inds_flat, shape))
    inds_copy = np.copy(inds)

    shuffle = np.random.permutation(inds.shape[1])
    inds2 = inds[:, shuffle]

    seed = 179
    arr1 = inds_to_normal(inds, shape, 0, size, seed=seed).T
    arr2 = inds_to_normal(inds2, shape, 0, size, seed=seed).T
    assert np.allclose(arr1[:, shuffle], arr2)

    arr3 = inds_to_normal(inds, shape, 0, size, seed=seed + 1).T
    assert not np.allclose(arr1, arr3)

    assert arr1.shape == (size, np.prod(shape))

    # Check if `inds` was changed
    assert np.allclose(inds, inds_copy)


def test_bit_randomness():
    """Test if most bits of hashed_int_c are pretty random"""
    N = 10000
    hashed = np.arange(10000, dtype=np.uint64)
    hash_int_c(hashed)
    hashed_bits = eight_bytes_to_bits(hashed)
    average_deviation = np.mean(hashed_bits, axis=0)
    average_deviation = np.abs(average_deviation - 0.5)
    average_deviation = np.mean(average_deviation)
    # roughly equal chance of being 0 or 1
    assert average_deviation < 0.06

    # roughly 50/50 chance of being different from neighbor
    average_deviation2 = np.mean(
        np.abs(np.diff(hashed_bits.astype(float), axis=1)), axis=0
    )
    average_deviation2 = np.abs(average_deviation2 - 0.5)
    average_deviation2 = np.mean(average_deviation2)
    assert average_deviation2 < 0.06


def test_inds_to_sparse_sign():
    shape = (10, 12, 14, 7)
    tot_dim = np.prod(shape, dtype=int)
    N = tot_dim
    np.random.seed(179)
    inds_flat = np.random.choice(tot_dim, size=N)
    inds_flat = np.arange(tot_dim)
    inds = np.array(np.unravel_index(inds_flat, shape))
    seed = 5
    rank = 100
    nonzero_per_row = 20

    nums = np.array(
        inds_to_sparse_sign(inds, shape, rank, 0, rank, nonzero_per_row, seed)
    )

    assert nums.shape == (tot_dim, rank)

    # Same chance of having non-zero element in each column
    mean_deviation = np.mean(
        np.abs(np.mean(np.abs(nums), axis=0) - nonzero_per_row / rank)
    )
    assert mean_deviation < 0.01

    # number of nonzero per row is correct
    assert np.all(np.sum(np.abs(nums), axis=1) == nonzero_per_row)

    # only +/-1
    assert np.all(np.unique(nums) == np.array([-1, 0, 1]))

    rank_min = 5
    rank_max = 20
    nums2 = np.array(
        inds_to_sparse_sign(
            inds, shape, rank, rank_min, rank_max, nonzero_per_row, seed
        )
    )
    assert np.all(nums[:, rank_min:rank_max] == nums2)


def test_inds_to_rand_double():
    """Test uniform distribution of mantissa and exponent"""
    shape = (10, 12, 14, 7)
    rank = 17
    tot_dim = np.prod(shape, dtype=int)
    N = tot_dim
    inds_flat = np.random.choice(tot_dim, size=N)
    inds_flat = np.arange(tot_dim)
    inds = np.array(np.unravel_index(inds_flat, shape))
    inds = inds.astype(np.uint64)
    seed = 5
    shape_arr = np.array(shape, dtype=np.uint64)
    inds_copy = np.copy(inds)

    rank_min = rank - 5
    rand_nums = np.array(
        _inds_to_rand_double(inds_copy, shape_arr, 0, rank, seed)
    )
    assert np.all(inds == inds_copy)  # function is not allowed to edit inds

    # Test rand_min
    rand_nums2 = np.array(
        _inds_to_rand_double(inds, shape_arr, rank_min, rank, seed)
    ).reshape(N, rank - rank_min)
    assert np.all(
        rand_nums.reshape(N, rank)[:, rank_min:] == rand_nums2
    )  # function
    mant, exp = np.frexp(rand_nums)

    # Check mantissa is always in [0.5,1]
    assert np.min(mant) >= 0.5
    assert np.max(mant) <= 1

    # Check the exponent is in the expected range
    assert np.min(exp) >= -510
    assert np.max(exp) <= 1

    # some statistical tests
    uniform_dev = np.sort(mant) - np.linspace(0.5, 1, tot_dim * rank)
    assert np.mean(np.abs(uniform_dev)) < 0.001
    assert (
        np.mean(np.abs(np.sort(exp) - np.linspace(-510, 1, tot_dim * rank))) < 2
    )

    # Test if increasing rank leaves first columns invariant
    rank2 = rank + 5
    rand_nums1 = np.reshape(rand_nums, (N, rank))
    rand_nums2 = np.array(
        _inds_to_rand_double(inds_copy, shape_arr, 0, rank2, seed)
    )
    rand_nums2 = np.reshape(rand_nums2, (N, rank2))
    assert np.all(rand_nums2[:, :rank] == rand_nums1)


def test_inds_to_normal():
    shape = (10, 12, 14, 7)
    rank = 17
    tot_dim = np.prod(shape, dtype=int)
    N = tot_dim
    inds_flat = np.random.choice(tot_dim, size=N)
    inds_flat = np.arange(tot_dim)
    inds = np.array(np.unravel_index(inds_flat, shape))
    inds = inds.astype(np.uint64)
    seed = 5
    shape_arr = np.array(shape, dtype=np.uint64)
    rand_nums = np.array(inds_to_normal(inds, shape_arr, 0, rank, seed))
    dist = np.sort(scipy.stats.norm.cdf(rand_nums.reshape(-1)))
    deviation = np.mean(np.abs(dist - np.linspace(0, 1, len(dist))))
    assert deviation < 0.01

    rank_min = 5
    rand_nums1 = rand_nums.reshape(N, rank)
    rand_nums2 = np.array(
        inds_to_normal(inds, shape_arr, rank_min, rank, seed)
    ).reshape(N, rank - rank_min)
    assert np.all(rand_nums1[:,rank_min:] == rand_nums2)



# %%
from itertools import product
import numpy as np
from tqdm import tqdm
from tt_sketch.tensor_train_sketching_matrix import TensorTrainDRM
from tt_sketch.tensor import dense_to_sparse
from tt_sketch.lazy_gaussian_sketching_matrix import GaussianDRM
from tt_sketch.sketched_tensor_train import (
    SketchedTensorTrain,
)
from tt_sketch.tt_svd import tt_svd
from tt_sketch.recursive_big_sketch import recursive_big_sketch
from tt_sketch.tensor import TensorTrain

# Shape of problem
n_dim = 4
dim = 20
n_samples = 1
shape = (dim,) * n_dim
seed = 179
print(f"{shape=}, total dim={np.prod(shape):.5e}")

# Store results in dictionary
results = dict()
rank_range = list(range(2, dim))
i_range = np.arange(n_samples)
methods = ["classical", "gauss-gauss", "tt-gauss", "tt-tt", "recursive"]
for name in methods:
    results[name] = np.ones((np.max(rank_range) + 1, len(i_range))) * np.nan


# Construct the dense tensor
X = TensorTrain.random(shape, 10)
X = X.to_numpy()

for r, i in tqdm(list(product(rank_range, i_range))):
    tt = tt_svd(X, r)
    error_tt = np.linalg.norm(tt.to_numpy() - X)
    results["classical"][r, i] = error_tt

    X_sparse = dense_to_sparse(X)

    left_rank = (r,) * (len(shape) - 1)
    right_rank = (r * 2 + 1,) * (len(shape) - 1)
    tt_sketched = stream_sketch(
        X_sparse,
        left_rank,
        right_rank,
        left_sketch_type=GaussianDRM,
        right_sketch_type=GaussianDRM,
    )
    tt2 = TensorTrain(tt_sketched.C_cores)
    error = np.linalg.norm(tt2.to_numpy() - X)
    results["gauss-gauss"][r, i] = error

    tt_sketched = stream_sketch(
        X_sparse,
        left_rank,
        right_rank,
        left_sketch_type=TensorTrainDRM,
        right_sketch_type=GaussianDRM,
    )
    tt2 = TensorTrain(tt_sketched.C_cores)
    error = np.linalg.norm(tt2.to_numpy() - X)
    results["tt-gauss"][r, i] = error

    tt_sketched = stream_sketch(
        X_sparse,
        left_rank,
        right_rank,
        left_sketch_type=TensorTrainDRM,
        right_sketch_type=TensorTrainDRM,
    )
    tt2 = TensorTrain(tt_sketched.C_cores)
    error = np.linalg.norm(tt2.to_numpy() - X)
    results["tt-tt"][r, i] = error

    cores = recursive_big_sketch(X, left_rank, right_rank)
    tt3 = TensorTrain(cores)
    error = np.linalg.norm(tt3.to_numpy() - X)
    results["recursive"][r, i] = error
# %%
import matplotlib.pyplot as plt

plt.figure(figsize=(10, 6))
for i, name in enumerate(methods):
    A = results[name]
    A = A[min(rank_range) :, :]
    A_medians = np.median(A, axis=1)
    A_20 = np.percentile(A, 20, axis=1)
    A_80 = np.percentile(A, 80, axis=1)
    plt.fill_between(rank_range, A_20, A_80, alpha=0.2)
    plt.plot(rank_range, A_medians, "-o", label=name)
plt.ylabel("Error (l2)")
plt.xlabel("tt-rank (uniform)")
plt.title(
    f"Big sketch (oversampling=2x) reconstruction error on {shape} random TT of rank 10"
)
plt.legend()
plt.yscale("log")
plt.savefig("exact_recovery.pdf", bbox_inches="tight")



"""
Implements methods for dispatching sketching methods for tensors and DRMs.
"""

import enum
from functools import partial
from typing import Callable, Literal, Optional, Tuple

import numpy as np
import numpy.typing as npt
import scipy.linalg

from tt_sketch.drm import TensorTrainDRM
from tt_sketch.drm_base import DRM
from tt_sketch.sketch_container import SketchContainer
from tt_sketch.sketching_methods.abstract_methods import (
    CansketchCP,
    CansketchDense,
    CansketchSparse,
    CansketchTT,
    CanSketchTucker,
)
from tt_sketch.sketching_methods.cp_sketch import sketch_omega_cp, sketch_psi_cp
from tt_sketch.sketching_methods.dense_sketch import (
    sketch_omega_dense,
    sketch_psi_dense,
)
from tt_sketch.sketching_methods.sparse_sketch import (
    sketch_omega_sparse,
    sketch_psi_sparse,
)
from tt_sketch.sketching_methods.tensor_train_sketch import (
    sketch_omega_tt,
    sketch_psi_tt,
)
from tt_sketch.sketching_methods.tucker_sketch import (
    sketch_omega_tucker,
    sketch_psi_tucker,
)
from tt_sketch.tensor import (
    CPTensor,
    DenseTensor,
    SparseTensor,
    Tensor,
    TensorSum,
    TensorTrain,
    TuckerTensor,
)
from tt_sketch.utils import ArrayList, right_mul_pinv

ABSTRACT_TENSOR_SKETCH_DISPATCH = {
    SparseTensor: CansketchSparse,
    TensorTrain: CansketchTT,
    DenseTensor: CansketchDense,
    CPTensor: CansketchCP,
    TuckerTensor: CanSketchTucker,
}

DRM_SKETCH_METHOD_DISPATCH = {
    SparseTensor: "sketch_sparse",
    TensorTrain: "sketch_tt",
    DenseTensor: "sketch_dense",
    CPTensor: "sketch_cp",
    TuckerTensor: "sketch_tucker",
}


OMEGA_METHODS = {
    SparseTensor: sketch_omega_sparse,
    TensorTrain: sketch_omega_tt,
    DenseTensor: sketch_omega_dense,
    CPTensor: sketch_omega_cp,
    TuckerTensor: sketch_omega_tucker,
}

PSI_METHODS = {
    SparseTensor: sketch_psi_sparse,
    TensorTrain: sketch_psi_tt,
    DenseTensor: sketch_psi_dense,
    CPTensor: sketch_psi_cp,
    TuckerTensor: sketch_psi_tucker,
}


def sketch_omega_sum(
    left_sketch_array: ArrayList,
    right_sketch_array: ArrayList,
    *,
    tensor: TensorSum,
    omega_shape: Tuple[int, int],
    **kwargs,
) -> npt.NDArray:
    omega = np.zeros(omega_shape)
    for summand, left_sketch, right_sketch in zip(
        tensor.tensors, left_sketch_array, right_sketch_array
    ):
        omega_method = OMEGA_METHODS[type(summand)]
        omega += omega_method(
            left_sketch,
            right_sketch,
            tensor=summand,
            omega_shape=omega_shape,
            **kwargs,
        )  # type: ignore
    return omega


OMEGA_METHODS[TensorSum] = sketch_omega_sum


def sketch_psi_sum(
    left_sketch_array: ArrayList,
    right_sketch_array: ArrayList,
    *,
    tensor: TensorSum,
    psi_shape: Tuple[int, int],
    **kwargs,
) -> npt.NDArray:
    psi = np.zeros(psi_shape)
    if left_sketch_array is None:
        left_sketch_array = (None,) * tensor.num_summands
    if right_sketch_array is None:
        right_sketch_array = (None,) * tensor.num_summands

    for summand, left_sketch, right_sketch in zip(
        tensor.tensors, left_sketch_array, right_sketch_array
    ):
        psi_method = PSI_METHODS[type(summand)]
        psi += psi_method(
            left_sketch,
            right_sketch,
            tensor=summand,
            psi_shape=psi_shape,
            **kwargs,
        )  # type: ignore
    return psi


PSI_METHODS[TensorSum] = sketch_psi_sum


def sum_sketch(tensor: TensorSum, *, drm: DRM):
    sketch_generators = []
    for summand in tensor.tensors:
        sketch_generators.append(get_sketch_method(summand, drm)(summand))
    for _ in range(len(tensor.shape) - 1):
        yield tuple(next(gen) for gen in sketch_generators)


def get_sketch_method(tensor: Tensor, drm: DRM) -> Callable:
    if type(tensor) in DRM_SKETCH_METHOD_DISPATCH:
        drm_sketch_method = DRM_SKETCH_METHOD_DISPATCH[type(tensor)]
        return getattr(drm, drm_sketch_method)
    elif isinstance(tensor, TensorSum):
        return partial(sum_sketch, drm=drm)
    else:
        raise ValueError(f"DRM of type {type(drm)} can't sketch {type(tensor)}")


def orth_step(
    Psi: npt.NDArray[np.float64], Omega: Optional[npt.NDArray[np.float64]]
) -> npt.NDArray[np.float64]:
    """
    Perform the orthogonalization step in the orthogonal sketching algorithm.
    """
    Psi_shape = Psi.shape
    final_right_rank = Psi_shape[2] if Omega is None else Omega.shape[0]
    Psi_mat = Psi.reshape((Psi_shape[0] * Psi_shape[1], Psi_shape[2]))
    if Omega is not None:
        Psi_mat = right_mul_pinv(Psi_mat, Omega)
    # Psi_mat, _ = np.linalg.qr(Psi_mat)
    Psi_mat, _ = scipy.linalg.qr(Psi_mat, mode="economic")
    Psi = Psi_mat.reshape(Psi_shape[0], Psi_shape[1], final_right_rank)
    return Psi


class OrthogTTDRM:
    """Represents the orthogonalized TT used as left-sketch for psi"""

    def __init__(self, rank, tensor):
        self.rank = rank
        self.drm = TensorTrainDRM(rank, tensor.shape, transpose=False, cores=[])
        self.generator = None
        self.tensor = tensor
        self.sketch_method = get_sketch_method(tensor, self.drm)

    def add_core(self, core):
        self.drm.cores.append(core)
        if self.generator is None:
            self.generator = self.sketch_method(self.tensor)

    def __next__(self):
        return next(self.generator)


class SketchMethod(enum.Enum):
    streaming = "streaming"
    orthogonal = "orthogonal"
    hmt = "hmt"


def general_sketch(
    tensor: Tensor,
    left_drm: Optional[DRM],
    right_drm: DRM,
    method: SketchMethod,
) -> SketchContainer:
    """General algorithm for sketching a tensor.

    Does the heavy lifting for both the streaming and orthogonal sketching
    algorithms."""
    n_dims = len(tensor.shape)

    if method != SketchMethod.hmt:
        if left_drm is None:
            raise ValueError(f"left_drm must be provided for method '{method}'")
        left_contractions = list(get_sketch_method(tensor, left_drm)(tensor))
    right_contractions = list(get_sketch_method(tensor, right_drm)(tensor))

    if left_drm is None:
        # This is just required for shape information in the case of HMT
        left_drm = right_drm.T

    Psi_cores: ArrayList = []

    # Compute Omega matrices
    Omega_mats: ArrayList = []
    if method != SketchMethod.hmt:
        omega_method = OMEGA_METHODS[type(tensor)]
        for mu in range(n_dims - 1):
            omega_shape = (left_drm.rank[mu], right_drm.rank[::-1][mu])
            Omega_mats.append(
                omega_method(
                    left_contractions[mu],
                    right_contractions[mu],
                    tensor=tensor,
                    mu=mu,
                    omega_shape=omega_shape,
                )  # type: ignore
            )

    if method in (SketchMethod.hmt, SketchMethod.orthogonal):
        left_psi_drm = OrthogTTDRM(left_drm.rank, tensor)

    # Compute Psi cores
    psi_method = PSI_METHODS[type(tensor)]
    for mu in range(n_dims):
        if mu > 0:
            if method in (SketchMethod.hmt, SketchMethod.orthogonal):
                left_psi_drm.add_core(Psi_cores[-1])
                left_sketch = next(left_psi_drm)
            else:
                left_sketch = left_contractions[mu - 1]
            r1 = left_drm.rank[mu - 1]
        else:
            left_sketch = None
            r1 = 1
        if mu < n_dims - 1:
            right_sketch = right_contractions[mu]
            r2 = right_drm.rank[::-1][mu]
        else:
            right_sketch = None
            r2 = 1

        psi_shape = (r1, tensor.shape[mu], r2)
        Psi = psi_method(
            left_sketch, right_sketch, tensor=tensor, mu=mu, psi_shape=psi_shape
        )  # type: ignore
        if mu < n_dims - 1:
            if method == SketchMethod.orthogonal:
                Psi = orth_step(Psi, Omega_mats[mu])
            elif method == SketchMethod.hmt:
                Psi = orth_step(Psi, None)
        Psi_cores.append(Psi)
    return SketchContainer(Psi_cores, Omega_mats)
